import pyqtgraph as pg
from PyQt6 import QtCore


class _KeyPressGraphicsLayoutWidget(pg.GraphicsLayoutWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    keyPressed = QtCore.pyqtSignal(object)
    keyReleased = QtCore.pyqtSignal(object)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        self.keyPressed.emit(event)

    def keyReleaseEvent(self, event):
        super().keyReleaseEvent(event)
        self.keyReleased.emit(event)



"""Functions to load data from the APE beamline at Elettra."""

import pint_xarray

from peaks.core.fileIO.base_arpes_data_classes.base_ses_class import SESDataLoader
from peaks.core.fileIO.loc_registry import register_loader

ureg = pint_xarray.unit_registry


@register_loader
class APEArpesLoader(SESDataLoader):
    """Data loader for APE ARPES system at Elettra.

    Notes
    ------------
    Loading spin data is not yet supported.

    """

    _loc_name = "Elettra_APE_LE"
    _loc_description = "LE (SARPES) branch of APE beamline at Elettra"
    _loc_url = "https://www.elettra.eu/elettra-beamlines/ape.html"
    _analyser_slit_angle = 0 * ureg("deg")

    _manipulator_name_conventions = {
        "polar": "polar",
        "tilt": "tilt",
        "x1": "X",
        "x2": "Z",
        "x3": "Y",
    }
    _SES_metadata_units = {
        f"manipulator_{dim}": ("mm" if dim in ["x1", "x2", "x3"] else "deg")
        for dim in _manipulator_name_conventions.keys()
    }



import functools
import re

import xarray as xr
from termcolor import colored

from peaks.core.utils.misc import analysis_warning


# Helper functions
def _ensure_empty_node(func):
    """Decorator to check if the DataTree node is empty before calling the function."""

    @functools.wraps(func)
    def wrapper(dt, *args, **kwargs):
        if not dt.is_empty:
            raise ValueError(
                f"DataTree node already appears to contain data. "
                f"`.{func.__name__}` should only be called on an empty node of the DataTree"
            )
        return func(dt, *args, **kwargs)

    return wrapper


def _ensure_name_does_not_exist(func):
    """Decorator to enforce passed name does not exist in the DataTree before calling the function,
    to avoid overwriting existing entries."""

    @functools.wraps(func)
    def wrapper(dt, *args, **kwargs):
        if kwargs.get("name") and kwargs.get("name") in dt:
            raise ValueError(
                f"Name '{kwargs.get('name')}' already exists as a child node in the DataTree. Please provide a unique name."
            )
        return func(dt, *args, **kwargs)

    return wrapper


def _make_name_callable(name):
    """Replace any character that is not a letter, digit, or underscore with an underscore"""
    return re.sub(r"[^a-zA-Z0-9_]", "_", name)


def _make_name_unique(name, dt):
    """Make a name unique by appending a number to it.

    Parameters
    ----------
    name : str
        Name to make unique.

    dt : DataTree
        DataTree to check for name uniqueness.

    Returns
    -------
    str
        Unique name.
    """
    index = 0
    base_name = name
    name = f"{base_name}_{index}"
    while name in dt:
        index += 1
        name = f"{base_name}_{index}"
    return name


def _dataarrays_to_datatree(data, names=None):
    """Parse one or a `list` of :class:`xarray.DataArray`'s into a :class:`xarray.DataTree` structure

    Parameters
    ----------
    data : xarray.DataArray or list of xarray.DataArray
        The data to be parsed into a DataTree structure.

    names : list of str, optional
        Names to assign to each DataArray. If not provided, the names will be automatically generated.
    """

    # Parse the names automatically if the names list not provided or not suitable
    if names and (len(names) != len(data) or len(set(names)) != len(names)):
        analysis_warning(
            "The names provided are not suitable for the number of files loaded. Pass a list of unique strings for "
            "each file loaded. Automatically generating names.",
            title="Loading info",
            warn_type="info",
        )
        names = None
    if not names:
        names = [_make_name_callable(da.name) for da in data]

    ds_dict = {
        da_name: (
            da.to_dataset(name="data", promote_attrs=False)
            if isinstance(da, xr.DataArray)
            else da  # Pass a DataSet directly if it is already a DataSet
        )
        for da_name, da in zip(names, data, strict=True)
    }

    return xr.DataTree.from_dict(ds_dict)


# General accessor methods for DataTree
def view(dt):
    """Make a nice view of all branches and scans in the DataTree.

    Parameters
    ----------
    dt : DataTree
        The DataTree to list scans from.
    """

    # Function to recursively iterate over DataTree and display in a nice structure
    def display_datatree(node, level=0, is_last=True):
        indent = ""

        if level > 0:
            if is_last:
                indent = "└── "
            else:
                indent = "├── "

        if node.name:
            name = node.name
        else:
            name = ""

        # Color the name if the node has a dataset
        if not node.is_empty:
            name = colored(name, "green")

        output = f"{'   ' * (level - 1) + indent}{name}\n"

        # Iterate over children, marking last child appropriately
        children = list(node.children.values())
        for i, child in enumerate(children):
            output += display_datatree(child, level + 1, i == len(children) - 1)

        return output

    # Display the DataTree structure
    print(display_datatree(dt))


@_ensure_empty_node
@_ensure_name_does_not_exist
def add_scan_group(dt, name=None):
    """Add a new group (branch) to the DataTree.

    Parameters
    ----------
    dt : DataTree
        The DataTree to add the group to.

    name : str, optional
        Name of the group to be added to the DataTree. Defaults to `scan_group_#` where # is a number to make the
        name unique.

    Examples
    --------
    Example usage is as follows::

        import peaks as pks

        # Create a new DataTree
        dt = xr.DataTree()

        # Add a new group to the DataTree
        dt.add_scan_group()
    """

    if not name:
        name = _make_name_unique("scan_group", dt)
    new_name = _make_name_callable(name)
    new_dt = xr.DataTree(name=new_name)
    dt[new_name] = new_dt


@_ensure_empty_node
@_ensure_name_does_not_exist
def add(dt, data_source, name=None, add_at_root=False, **kwargs):
    """Add data into an existing DataTree. Can be either loaded data, or arguments to pass to the main `peaks.load`
    function.

    Parameters
    ----------
    dt : DataTree
        The DataTree to add data into.

    data_source : str or int or xarray.DataArray or xarray.Dataset or xarray.DataTree
        File path identifier to load data from, or data to add to the DataTree.

    name : str, optional
        Name of the branch to be added to the :class:`xarray.DataTree`. Defaults to `scan_group_#` if a group of scans
        are added, or the name of the file if a single scan is added. Note, if loading multuple scans, the `names`
        argument can also be provided to specify the names of each scan within the added scan group.

    add_at_root : bool, optional
        If `True`, and if a :class:`xarray.DataTree` is passed, the children of the added :class:`xarray.DataTree`
         will all be be added at the root level of the original :class:`xarray.DataTree`. Defaults to `False`, in which
         case the new :class:`xarray.DataTree` is added as a new branch of the tree, with the passed name or an
         automatically generated name. If set `True`, the `name` argument is ignored.

    **kwargs : other keyword arguments to pass to `peaks.core.fileIO.data_loading.load`

    See Also
    --------
    peaks.core.fileIO.data_loading.load : The function that is called to load the data.

    Examples
    --------
    Example usage is as follows::

        import peaks as pks

        # Create a new DataTree by loading a set of data
        disps = pks.load(['disp1.ibw','disp2.ibw'])

        # Add a new dispersion into the DataTree
        disps.add('disp3.ibw')
    """

    if isinstance(data_source, xr.DataTree):
        # Group of scans passed.
        if add_at_root:
            # Add children of the DataTree to the root of the original DataTree
            for child in data_source.children.values():
                dt[child.name] = child
            return

        # Otherwise, add them as a new group with the given name or a default name
        name = _make_name_callable(name) if name else _make_name_unique("scan_group", dt)
        data_source.name = name
        dt[name] = data_source
        return

    if isinstance(data_source, xr.DataArray):
        # Single da passed - parse it to DataTree format and add to the tree
        data = _dataarrays_to_datatree([data_source], names=[name] if name else None)
        data = data[list(data)[0]]
        dt[data.name] = data
        return

    if isinstance(data_source, xr.Dataset):
        # Single ds passed - parse it to DataTree format and add to the tree
        if not name:
            raise ValueError(
                "A name must be provided when adding a Dataset to the DataTree."
            )
        data = xr.DataTree(name=_make_name_callable(name), dataset=data_source)
        dt[data.name] = data
        return

    if isinstance(data_source, (str, int, list)):
        from peaks.core.fileIO.data_loading import load as pks_load

        # This should be a file path or a file identifier - load the data
        loaded_data = pks_load(data_source, **kwargs)
        # Add loaded data to the DataTree
        add(dt=dt, data_source=loaded_data, name=name, add_at_root=add_at_root)


def get_DataArray(dt):
    """Get a :class:`xarray.DataArray` stored as the only entry of a :class:`xarray.DataSet` with
     data variable `data` which is the leaf node of a :class:`xarray.DataTree`.

    Parameters
    ----------
    dt : DataTree leaf
        The DataTree leaf to return the :class:`xarray.DataArray` from.

    Returns
    -------
    xarray.DataArray
        The :class:`xarray.DataArray` stored in the leaf node.
    """
    if not dt.is_empty and hasattr(dt, "data") and isinstance(dt.data, xr.DataArray):
        da = dt.data
        da.name = dt.name
        return da


def get_list_of_DataArrays_from_DataTree(dt):
    """Get a list of :class:`xarray.DataArray` stored as the only entry of a :class:`xarray.DataSet` with
     data variable `data` which are the leaf nodes of a :class:`xarray.DataTree`.

    Parameters
    ----------
    dt : DataTree
        The DataTree to return the :class:`xarray.DataArray` from.

    Returns
    -------
    list of xarray.DataArray
        The :class:`xarray.DataArray`s stored in the leaf nodes.
    """
    data_list = []
    for node in dt.subtree:
        if not node.is_empty:
            data_list.append(get_DataArray(node))
    return data_list


def _map_over_dt_containing_single_das(data_tree, func, *args, **kwargs):
    """Applies a function to each leaf node in a DataTree assuming that any leaf with data contains a single
    :class:`xarray.Dataset` with a single data variable `data` containing the :class:`xarray.DataArray`.

    Parameters
    ----------
    data_tree : xarray.DataTree
        The DataTree to apply the function to.
    func : callable
        The function to apply to each leaf node.
    *args, **kwargs
        Additional arguments to pass to the function.

    Returns
    -------
    xarray.DataTree
        The DataTree with the function applied to each leaf node.
    """

    def map_over_ds(node):
        if hasattr(node, "data") and len(node) == 1:
            return node.map(lambda da: func(da, *args, **kwargs) or da)
        else:
            return node

    return data_tree.map_over_datasets(map_over_ds)



import os

import pandas as pd
from bokeh.io import output_notebook
from bokeh.plotting import figure, show
from bokeh.transform import dodge, factor_cmap


def ptab(modify_elements_func=None, add_to_tooltips=None):
    """Plot the periodic table
    Modified from https://docs.bokeh.org/en/2.4.3/docs/gallery/periodic.html

    Parameters
    ----------
    modify_elements_func : callable, optional
        Function to modify the elements data frame. Should accept a single argument of
        the elements data frame and return a modified data frame.
    add_to_tooltips : tuple | list of tuples, optional
        Additional tooltips to add to the plot.
    """
    output_notebook()

    # Load the elements data
    # elements.pkl is a DataFrame bundled from bokeh.sample_data, License: Public Domain
    elements_path = os.path.join(os.path.dirname(__file__), "elements.pkl")
    elements = pd.read_pickle(elements_path)

    # Define period and group information
    periods = ["I", "II", "III", "IV", "V", "VI", "VII", "", "*", "**"]
    groups = [str(x) for x in range(1, 19)]

    # Create Data frame for main group elements
    global df
    df = elements.copy()
    df["atomic mass"] = df["atomic mass"].astype(str)
    df["period"] = [periods[x - 1] for x in df.period]

    # Modify group and period information of LA and AC series for plotting purposes
    for ii in range(56, 71):
        df.loc[ii, "period"] = "*"
        df.loc[ii, "group"] = 4 + ii - 56
    for ii in range(88, 103):
        df.loc[ii, "period"] = "**"
        df.loc[ii, "group"] = 4 + ii - 88
    df["group"] = df["group"].astype(str)
    df.loc[70, "metal"] = "lanthanoid"
    df.loc[102, "metal"] = "actinoid"

    # Correct name for Hn
    df.loc[112, "name"] = "Nihonium"

    # Modify the elements data frame if necessary
    if modify_elements_func:
        df = modify_elements_func(df)

    cmap = {
        "alkali metal": "#a6cee3",
        "alkaline earth metal": "#1f78b4",
        "metal": "#d93b43",
        "halogen": "#999d9a",
        "metalloid": "#e08d49",
        "noble gas": "#eaeaea",
        "nonmetal": "#f1d4Af",
        "transition metal": "#65C99C",
        "lanthanoid": "#FFFFE0",
        "actinoid": "#FFD9D9",
    }

    TOOLTIPS = [
        ("Name", "@name"),
        ("Atomic number", "@{atomic number}"),
        ("Atomic mass", "@{atomic mass}"),
        ("Type", "@metal"),
        ("Electronic configuration", "@{electronic configuration}"),
    ]
    # Add additional tooltips if necessary
    if add_to_tooltips:
        if isinstance(add_to_tooltips, tuple):
            TOOLTIPS.append(add_to_tooltips)
        else:
            TOOLTIPS.extend(add_to_tooltips)

    p = figure(
        title="Periodic Table",
        width=1000,
        height=550,
        x_range=groups,
        y_range=list(reversed(periods)),
        tools="hover",
        toolbar_location=None,
        tooltips=TOOLTIPS,
    )

    r = p.rect(
        "group",
        "period",
        0.95,
        0.95,
        source=df,
        fill_alpha=0.6,
        legend_field="metal",
        color=factor_cmap(
            "metal", palette=list(cmap.values()), factors=list(cmap.keys())
        ),
    )

    text_props = dict(source=df, text_align="left", text_baseline="middle")

    x = dodge("group", -0.4, range=p.x_range)

    p.text(x=x, y="period", text="symbol", text_font_style="bold", **text_props)

    p.text(
        x=x,
        y=dodge("period", 0.3, range=p.y_range),
        text="atomic number",
        text_font_size="11px",
        **text_props,
    )

    p.text(
        x=x,
        y=dodge("period", -0.35, range=p.y_range),
        text="name",
        text_font_size="7px",
        **text_props,
    )

    p.text(
        x=x,
        y=dodge("period", -0.2, range=p.y_range),
        text="atomic mass",
        text_font_size="7px",
        **text_props,
    )

    p.text(
        x=["3", "3"],
        y=["VI", "VII"],
        text=["LA", "AC"],
        text_align="center",
        text_baseline="middle",
    )

    p.outline_line_color = None
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_standoff = 0
    p.legend.orientation = "horizontal"
    p.legend.location = (0, 105)
    p.hover.renderers = [r]

    show(p)



import os

import numpy as np
from igor2 import binarywave

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.fileIO.loc_registry import register_loader


@register_loader
class BaseIBWDataLoader(BaseDataLoader):
    """Base class for data loaders for Igor Binary Wave files."""

    _loc_name = "ibw"
    _loc_description = "General loader for Igor Binary Wave files"
    _loc_url = "https://www.wavemetrics.com/products"
    _metadata_parsers = ["_parse_wavenote_metadata"]

    @classmethod
    def _load_data(cls, fpath, lazy):
        """Load the data from an Igor Binary Wave file."""

        # Open the file and load its contents
        file_contents = binarywave.load(fpath)

        # Extract spectrum
        spectrum = file_contents["wave"]["wData"]

        # Extract relevant information on the dimensions of the data
        dim_size = file_contents["wave"]["bin_header"]["dimEUnitsSize"]
        dim_units = file_contents["wave"]["dimension_units"].decode()

        # Extract scales of dimensions
        dim_start = file_contents["wave"]["wave_header"]["sfB"]  # Initial value
        dim_step = file_contents["wave"]["wave_header"]["sfA"]  # Step size
        dim_points = file_contents["wave"]["wave_header"]["nDim"]  # Number of points
        dim_end = dim_start + (dim_step * (dim_points - 1))

        # Loop through the dimensions, extract relevant dimension names and coordinates
        dims = []
        coords = {}
        counter = 0
        for i in range(spectrum.ndim):
            dim = dim_units[counter : counter + dim_size[i]]
            dims.append(dim)
            coords[dim] = np.linspace(
                dim_start[i], dim_end[i], dim_points[i], endpoint=False
            )
            counter += dim_size[i]

        return {"spectrum": spectrum, "dims": dims, "coords": coords, "units": {}}

    @classmethod
    def _load_metadata(cls, fpath):
        """Load metadata from an Igor Binary Wave file."""
        # Load just the wavenote
        # Define maximum number of dimensions
        max_dims = 4

        # Read the ibw bin header segment of the file (IBW version 2,5 only)
        with open(fpath, "rb") as f:
            # Determine file version and extract file information
            version = np.fromfile(f, dtype=np.dtype("int16"), count=1)[0]
            if version == 2:
                # The size of the WaveHeader2 data structure plus the wave data plus 16
                # bytes of padding.
                wfmSize = np.fromfile(f, dtype=np.dtype("uint32"), count=1)[0]
                # The size of the note text.
                noteSize = np.fromfile(f, dtype=np.dtype("uint32"), count=1)[0]
                # Reserved. Write zero. Ignore on read.
                pictSize = np.fromfile(f, dtype=np.dtype("uint32"), count=1)[0]  # noqa: F841
                # Checksum over this header and the wave header.
                checksum = np.fromfile(f, dtype=np.dtype("int16"), count=1)[0]  # noqa: F841
            elif version == 5:
                # Checksum over this header and the wave header.
                checksum = np.fromfile(f, dtype=np.dtype("short"), count=1)[0]  # noqa: F841
                # The size of the WaveHeader5 data structure plus the wave data.
                wfmSize = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]  # noqa: F841
                # The size of the dependency formula, if any.
                formulaSize = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]  # noqa: F841
                # The size of the note text.
                noteSize = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]
                # The size of optional extended data units.
                dataEUnitsSize = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]
                # The size of optional extended dimension units.
                dimEUnitsSize = np.fromfile(f, dtype=np.dtype("int32"), count=max_dims)
                # The size of optional dimension labels.
                dimLabelsSize = np.fromfile(f, dtype=np.dtype("int32"), count=4)
                # The size of string indices if this is a text wave.
                sIndicesSize = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]
                # Reserved. Write zero. Ignore on read.
                optionsSize1 = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]
                # Reserved. Write zero. Ignore on read.
                optionsSize2 = np.fromfile(f, dtype=np.dtype("int32"), count=1)[0]

        # Open the file and read the wavenote
        with open(fpath, "rb") as f:
            # Move the cursor to the end of the file
            f.seek(0, os.SEEK_END)
            # Get the current position of pointer
            pointer_location = f.tell()

            # Determine file version-dependent offset
            if version == 2:
                offset = noteSize
            elif version == 5:
                # Work out file location of wavenote
                offset = (
                    noteSize
                    + dataEUnitsSize.sum()
                    + dimEUnitsSize.sum()
                    + dimLabelsSize.sum()
                    + sIndicesSize.sum()
                    + optionsSize1.sum()
                    + optionsSize2.sum()
                )

            # Move the file pointer to the location pointed by pointer_location,
            # considering the offset
            f.seek(pointer_location - offset)
            # read that bytes/characters to determine the wavenote
            wavenote = f.read(offset).decode()
            # NB No not cache the wavenote, as other loaders rely on this method and
            # that screws things up

        return {"wavenote": wavenote}

    @classmethod
    def _parse_wavenote_metadata(cls, metadata_dict):
        """Parse metadata specific to the wavenote."""
        return {"wavenote": metadata_dict.get("wavenote")}, []



import re
import zipfile

import numpy as np
import pint_xarray

from peaks import BaseIBWDataLoader
from peaks.core.fileIO.base_arpes_data_classes.base_arpes_data_class import (
    BaseARPESDataLoader,
)
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.utils.misc import analysis_warning

ureg = pint_xarray.unit_registry


@register_loader
class SESDataLoader(BaseARPESDataLoader):
    """Generic data loader for SES data.

    Notes
    ------------
    This class is intended to be subclassed to provide specific loaders for different
    locs using SES data formats.

    Subclasses should define the `_SES_metadata_key_mappings` class variable to map any
    custom metadata fields or fixed values to the standard peaks metadata keys. See the
    docstrings for `_SES_metadata_dict_keys_to_peaks_keys` for more information.

    Subclasses should also define the `_SES_metadata_units` dictionary to map any fixed
    units that are not determined as part of the metadata loader.

    """

    _loc_name = "SES"
    _loc_description = "Loader for data acquired using the Scienta Omicron SES software."
    _loc_url = "https://scientaomicron.com"
    # Dictionary to overwrite or add to default metadata keys:
    _SES_metadata_key_mappings = {}  # mappings from SES metadata key to peaks key
    _SES_metadata_units = {}  # standard SES units
    _analyser_name_conventions = {
        "deflector_parallel": "ThetaX",
        "deflector_perp": "ThetaY",
    }

    @classmethod
    def _load_data(cls, fpath, lazy, **kwargs):
        """Load SES data from different file types."""
        handlers = {
            "txt": cls._load_from_txt,
            "zip": cls._load_from_zip,
            "ibw": cls._load_from_ibw,
        }
        ext = fpath.split(".")[-1]
        if ext not in handlers:
            raise ValueError(
                f"File extension {ext} is not supported for loader {cls.__name__}"
            )

        # Load data
        scan_no = kwargs.pop("scan_no", None)
        if ext == "zip" and scan_no is not None:
            return handlers[ext](fpath, scan_no=scan_no)
        else:
            return handlers[ext](fpath)

    @classmethod
    def _load_from_txt(cls, fpath):
        """Load data from a .txt file."""
        # Give a legacy warning
        analysis_warning(
            "Loading SES data from .txt file. This is a legacy format and only basic "
            "loading is supported. It is strongly encouraged to save data using the "
            ".ibw format from the SES software.",
            "warning",
            "Legacy format - SES .txt file loader",
        )

        # Get the metadata here returning with keys in SES format - this as needed for
        # parsing the core data.
        metadata_dict_ses_keys = cls._load_metadata(fpath, return_in_SES_format=True)
        # Cahce it in the metadata cache to avoid having to load it again later
        cls._metadata_cache[fpath] = metadata_dict_ses_keys

        # Load file data and core data-related metadata
        file_data = np.loadtxt(
            fpath, skiprows=int(metadata_dict_ses_keys["metadata_lines_length"])
        )
        spectrum = file_data[:, 1:]
        eV_values = file_data[:, 0]
        eV_units = cls._parse_SES_units_from_name(
            metadata_dict_ses_keys.get("Dimension 1 name", "")
        )
        theta_par_values = np.fromstring(
            metadata_dict_ses_keys["Dimension 2 scale"], sep=" "
        )
        theta_par_units = cls._parse_SES_units_from_name(
            metadata_dict_ses_keys.get("Dimension 2 name", "")
        )
        theta_par_label = "theta_par" if theta_par_units == "deg" else "y_scale"
        dims = ["eV", theta_par_label]
        coords = {"eV": eV_values, theta_par_label: theta_par_values}
        units = {"eV": eV_units, theta_par_label: theta_par_units, "spectrum": "counts"}

        return {
            "spectrum": spectrum,
            "dims": dims,
            "coords": coords,
            "units": units,
        }

    @classmethod
    def _load_from_zip(cls, fpath, scan_no=0):
        """Load data from standard SES .zip format.
        Adapted from the PESTO file loader by Craig Polley.

        Pass a `scan_no` to load a specific scan from the .zip file if multiple regions
        have been scanned together in SES."""
        # Open the file and load the data
        with zipfile.ZipFile(fpath) as z:
            files = z.namelist()
            file_bin = [file for file in files if ".bin" in file]
            file_ini = [file for file in files if "Spectrum_" in file and ".ini" in file]

            filename = file_bin[scan_no]
            filename_ini = file_ini[scan_no]

            # Extract coordinate information
            with z.open(filename_ini) as f:
                # Read and decode lines in file
                lines = f.readlines()
                lineText = [line.decode() for line in lines]

            # Convert relevant metadata to dictionary
            scan_metadata = cls._SES_metadata_to_dict_w_SES_keys(lineText)

            # Extract kinetic energy axis
            num_KE = int(scan_metadata["width"])
            KE_start = float(scan_metadata["widthoffset"])
            KE_step = float(scan_metadata["widthdelta"])
            KE_end = KE_start + (KE_step * (num_KE - 1))
            KE_values = np.linspace(KE_start, KE_end, num_KE)
            KE_units = cls._parse_SES_units_from_name(
                scan_metadata.get("widthlabel", "")
            )

            # Extract theta_par axis
            num_theta_par = int(scan_metadata["height"])
            theta_par_start = float(scan_metadata["heightoffset"])
            theta_par_step = float(scan_metadata["heightdelta"])
            theta_par_end = theta_par_start + (theta_par_step * (num_theta_par - 1))
            theta_par_values = np.linspace(theta_par_start, theta_par_end, num_theta_par)
            theta_par_units = cls._parse_SES_units_from_name(
                scan_metadata.get("heightlabel", "")
            )

            # Extract deflector axis
            num_defl_perp = int(scan_metadata["depth"])
            defl_perp_start = float(scan_metadata["depthoffset"])
            defl_perp_step = float(scan_metadata["depthdelta"])
            defl_perp_end = defl_perp_start + (defl_perp_step * (num_defl_perp - 1))
            defl_perp_values = np.linspace(defl_perp_start, defl_perp_end, num_defl_perp)
            defl_perp_units = cls._parse_SES_units_from_name(
                scan_metadata.get("depthlabel", "")
            )

            # Extract spectrum and reshape into a data cube to be consistent with loading
            with z.open(filename, "r") as f:
                spectrum = np.frombuffer(f.read(), dtype=np.dtype(np.float32))
                spectrum = spectrum.reshape(
                    num_KE, num_theta_par, num_defl_perp, order="F"
                )

        return {
            "spectrum": spectrum,
            "dims": ["eV", "theta_par", "deflector_perp"],
            "coords": {
                "eV": KE_values,
                "theta_par": theta_par_values,
                "deflector_perp": defl_perp_values,
            },
            "units": {
                "eV": KE_units,
                "theta_par": theta_par_units,
                "deflector_perp": defl_perp_units,
                "spectrum": "counts",
            },
        }

    @classmethod
    def _load_from_ibw(cls, fpath):
        """Load data from an Igor binary wave (ibw) file."""

        # Load the data from the ibw file using the default IBW loader
        data = BaseIBWDataLoader._load_data(fpath, lazy=False)

        # Load the metadata if needed for parsing the data
        pos_or_point_scan_dim = [
            axis
            for axis in data["dims"]
            if "position" in axis.lower() or "point" in axis.lower()
        ]
        has_binding_energy_dim = [
            axis for axis in data["dims"] if "binding" in axis.lower()
        ]
        has_hv_dim = [
            axis
            for axis in data["dims"]
            if "hv" in axis.lower() or "photon" in axis.lower()
        ]
        if has_binding_energy_dim or pos_or_point_scan_dim or has_hv_dim:
            metadata_dict_SES_keys = cls._load_metadata(fpath, return_in_SES_format=True)

            if pos_or_point_scan_dim:
                # Parse the manipulator (or other external) data for the scan

                # Extract the run mode information
                rmi = metadata_dict_SES_keys.get("Run Mode Information")
                # Remove the scan name line if present
                raw_manipulator_data = rmi[1:] if "Name" in rmi[0] else rmi

                # Parse the data into arrays
                split_data = [line.split("\x0b") for line in raw_manipulator_data]
                header = split_data[0]
                numeric_data = split_data[1:]
                numeric_array = np.array(numeric_data, dtype=float)
                manipulator_data = {
                    header[i]: numeric_array[:, i] for i in range(len(header))
                }

                # In the SES data format the Point and Position columns define the scan
                # and are duplicated in the raw axis columns
                # Check which arrays match
                matching_pairs = []
                keys = list(header)
                all_matched_keys = set()
                for i in range(len(keys)):
                    for j in range(i + 1, len(keys)):
                        key1 = keys[i]
                        key2 = keys[j]
                        if np.array_equal(
                            manipulator_data[key1], manipulator_data[key2]
                        ):
                            matching_pairs.append((key1, key2))
                            all_matched_keys.update([key1, key2])
                axis_mapping = {}
                for i in matching_pairs:
                    a, b = i
                    if "position" in a.lower() or "point" in a.lower():
                        axis_mapping[a] = b
                    else:
                        axis_mapping[b] = a

                inverse_manipulator_name_mapping = {
                    v: k
                    for k, v in cls._manipulator_name_conventions.items()
                    if v is not None
                }
                for axis in inverse_manipulator_name_mapping.keys():
                    # Check if it varies or is static
                    if abs(np.ptp(manipulator_data[axis])) < 1e-3:  # Same within noise
                        metadata_dict_SES_keys[axis] = manipulator_data[axis].mean()
                    else:
                        if axis in all_matched_keys:
                            # This is a scannable - just return min and max for some
                            # basic info in metadata
                            metadata_dict_SES_keys[axis] = np.asarray(
                                [
                                    np.min(manipulator_data[axis]),
                                    np.max(manipulator_data[axis]),
                                ]
                            )

                        else:
                            # This is likely an independent axis - return the full array
                            # to allow parsing if required
                            metadata_dict_SES_keys[axis] = manipulator_data[axis]

        # Handle some mapping of dimension names
        dim_new_name = data["dims"].copy()
        units = {}
        units["spectrum"] = "counts"
        for i, dim in enumerate(data["dims"]):
            if "kinetic" in dim.lower():
                dim_new_name[i] = "eV"
                units["eV"] = "eV"
            elif "binding" in dim.lower():
                if metadata_dict_SES_keys.get("Energy Unit") == "Kinetic":
                    if data["coords"][dim][0] > 0:  # Positive convention for BE
                        data["coords"][dim] = np.linspace(
                            float(metadata_dict_SES_keys.get("Low Energy")),
                            float(metadata_dict_SES_keys.get("High Energy")),
                            len(data["coords"][dim]),
                        )

                    else:
                        data["coords"][dim] = np.linspace(
                            float(metadata_dict_SES_keys.get("High Energy")),
                            float(metadata_dict_SES_keys.get("Low Energy")),
                            len(data["coords"][dim]),
                        )
                    dim_new_name[i] = "eV"
                    units["eV"] = "eV"
                    cls._SES_metadata_key_mappings["analyser_eV_type"] = lambda x: (
                        "Kinetic"
                    )
                else:
                    dim_new_name[i] = "eV"
                    units["eV"] = "eV"
                    cls._SES_metadata_key_mappings["analyser_eV_type"] = lambda x: (
                        "Binding"
                    )
                    analysis_warning(
                        "Data energy axis has been loaded as binding energy.",
                        title="Loading info",
                        warn_type="danger",
                    )
            elif "photon" in dim.lower() or "hv" in dim.lower():
                KE_delta = data["coords"][dim] - data["coords"][dim][0]
                units["hv"] = "eV"
                dim_new_name[i] = "hv"
                metadata_dict_SES_keys["Excitation Energy"] = [
                    data["coords"][dim][0],
                    data["coords"][dim][-1],
                ]
            elif "deg" in dim.lower():
                dim_new_name[i] = "theta_par"
                units["theta_par"] = "deg"
            elif "scale" in dim.lower():
                dim_new_name[i] = "y_scale"
                units["y_scale"] = cls._parse_SES_units_from_name(dim)
            elif "iteration" in dim.lower():
                dim_new_name[i] = "scan_no"
            elif "position" in dim.lower() or "point" in dim.lower():
                axis_name = axis_mapping[dim]
                peaks_axis_name = inverse_manipulator_name_mapping.get(
                    axis_name, axis_name
                )
                dim_new_name[i] = peaks_axis_name
                units[f"manipulator_{peaks_axis_name}"] = cls._SES_metadata_units.get(
                    peaks_axis_name
                )

        # Replace dimension names, co-ordinate labels, and units where required
        data["coords"] = {
            dim_new_name[i]: data["coords"][dim] for i, dim in enumerate(data["dims"])
        }
        if "hv" in data["coords"]:
            data["coords"]["KE_delta"] = ("hv", KE_delta)
            units["KE_delta"] = "eV"
        data["dims"] = dim_new_name
        data["units"].update(units)

        # Cache the metadata if it was loaded
        if "metadata_dict_SES_keys" in locals():
            cls._metadata_cache[fpath] = metadata_dict_SES_keys

        return data

    @classmethod
    def _load_metadata(cls, fpath, return_in_SES_format=False):
        """Load metadata from an SES file."""
        # Check for a cached version (only cache in this loader with keys in SES format)
        metadata_dict_SES_keys = cls._metadata_cache.get(fpath)

        # If no metadata in the cache, load it
        if not metadata_dict_SES_keys:
            handlers = {
                "txt": cls._load_SES_metadata_txt,
                "zip": cls._load_SES_metadata_zip,
                "ibw": cls._load_SES_metadata_ibw,
            }
            ext = fpath.split(".")[-1]
            metadata_lines = handlers[ext](fpath)
            metadata_dict_SES_keys = cls._SES_metadata_to_dict_w_SES_keys(metadata_lines)
            # Check if there is additional run mode information in the metadata
            try:
                run_mode_info_start_index = metadata_lines.index(
                    "[Run Mode Information]"
                )
                run_mode_info_stop_index = (
                    metadata_lines[run_mode_info_start_index:].index("")
                    + run_mode_info_start_index
                )
                metadata_dict_SES_keys["Run Mode Information"] = metadata_lines[
                    run_mode_info_start_index + 1 : run_mode_info_stop_index
                ]

            except ValueError:
                pass
        if return_in_SES_format:
            return metadata_dict_SES_keys

        # Convert the metadata to the peaks convention and return
        return cls._SES_metadata_dict_keys_to_peaks_keys(metadata_dict_SES_keys)

    @staticmethod
    def _load_SES_metadata_txt(fpath):
        """Extract the lines containing metadata in an SES format .txt file.

        Returns
        ------------
        metadata_lines : list
            Lines extracted from the file containing the metadata.

        """
        # Open the file and extract the lines containing metadata
        with open(fpath) as f:
            metadata_lines = []
            while True:
                line = f.readline()
                metadata_lines.append(line)
                # When the line starting with '[Data' is encountered, metadata has ended
                # and scan data has begun
                if "[Data" in line:
                    break
            # Manually add the lengths of the metadata section as needed for .txt loader
            metadata_lines.append(f"metadata_lines_length={len(metadata_lines)}\n")
        return metadata_lines

    @staticmethod
    def _load_SES_metadata_zip(fpath):
        """Extract the lines containing metadata in an SES format .zip file.

        Returns
        ------------
        metadata_lines : list
            Lines extracted from the file containing the metadata.

        """
        # Open the file and extract the lines containing metadata
        with zipfile.ZipFile(fpath, "r") as z:
            files = z.namelist()
            file_ini = [
                file for file in files if "Spectrum_" not in file and ".ini" in file
            ]

            if not file_ini:
                raise FileNotFoundError(
                    "No .ini file found in the .zip archive. Cannot extract metadata."
                )
            filename = file_ini[0]
            with z.open(filename, "r") as f:
                lines_to_decode = f.readlines()
                metadata_lines = [line.decode() for line in lines_to_decode]
        return metadata_lines

    @staticmethod
    def _load_SES_metadata_ibw(fpath):
        """Extract the lines containing metadata in an SES format .ibw file."""
        # Load wavenote using the BaseIBWDataLoader
        wavenote = BaseIBWDataLoader._load_metadata(fpath).get("wavenote")

        return wavenote.split("\r")

    @staticmethod
    def _parse_SES_units_from_name(name):
        """Parse the units from a key in the SES metadata."""
        return (
            re.search(r"\[(.*?)\]", name).group(1)
            if re.search(r"\[(.*?)\]", name)
            else None
        )

    @staticmethod
    def _SES_metadata_to_dict_w_SES_keys(metadata_lines):
        """Convert metadata lines to a dictionary of key-value pairs with the keys being
        the metadata entries as they appear in the SES metadata
        (i.e. not yet in :class:`peaks` convention).
        """

        return {
            line.split("=" if "=" in line else ":")[0].strip(): line.split(
                "=" if "=" in line else ":"
            )[1].strip()
            for line in metadata_lines
            if ("=" in line or ":" in line)
        }

    @classmethod
    def _SES_metadata_dict_keys_to_peaks_keys(cls, metadata_dict_SES_keys):
        """Extract metadata in peaks conventions and assign units where appropriate.

        Parameters
        ------------
        metadata_dict_SES_keys : dict
            Dictionary of metadata key-value pairs with keys in SES format.

        Returns
        ------------
        metadata_dict : dict
            Dictionary of metadata key-value pairs with keys in peaks format.

        Notes
        ------------
        The extraction process is based on the mappings defined in the dictionaries
        `standard_keys` and `standard_units`. The entries of these dictionaries are
        updated from the class variables `_SES_metadata_key_mappings` and
        `_SES_metadata_units` respectively, and so subclasses can overwrite and  extend
        these defaults by specifying the appropriate mappings in these class variables.

        The extraction process supports different types of keys:
            - If a single key is given, the value is extracted directly.
            - If a list of keys is given, the function tries to extract values for all
            keys and returns a list or array of all non-None values.
            - If a callable is given, the function calls it, passing it the metadata
            dictionary and returns the result.
            - A constant value can be given by defining a simple lambda function, for
            example `lambda x: -1`.

        """

        def _parse_timestamp(metadata_dict_SES_keys):
            """Parse the timestamp from the metadata."""
            # Extract the date and time from the metadata
            date = metadata_dict_SES_keys.get("Date")
            time = metadata_dict_SES_keys.get("Time")

            # If the date and time are present, combine them and return the timestamp
            if date and time:
                return f"{date} {time}"
            return None

        # Define the mapping assuming the standard keys for SES metadata
        standard_keys = {
            "analyser_model": "Instrument",
            "analyser_slit_width": None,
            "analyser_slit_width_identifier": None,
            "analyser_eV": ["Low Energy", "High Energy"],
            "analyser_eV_type": lambda x: x.get("Energy Unit") or x.get("Energy Scale"),
            "analyser_step_size": "Energy Step",
            "analyser_PE": "Pass Energy",
            "analyser_sweeps": "Number of Sweeps",
            "analyser_dwell": "Step Time",
            "analyser_lens_mode": "Lens Mode",
            "analyser_acquisition_mode": "Acquisition Mode",
            "analyser_polar": None,
            "analyser_tilt": None,
            "analyser_azi": None,
            "analyser_deflector_parallel": ["Thetax_Low", "Thetax_High", "ThetaX"],
            "analyser_deflector_perp": ["Thetay_Low", "Thetay_High", "ThetaY"],
            "timestamp": _parse_timestamp,
            "manipulator_polar": cls._manipulator_name_conventions.get("polar", None),
            "manipulator_tilt": cls._manipulator_name_conventions.get("tilt", None),
            "manipulator_azi": cls._manipulator_name_conventions.get("azi", None),
            "manipulator_x1": cls._manipulator_name_conventions.get("x1", None),
            "manipulator_x2": cls._manipulator_name_conventions.get("x2", None),
            "manipulator_x3": cls._manipulator_name_conventions.get("x3", None),
            "photon_hv": "Excitation Energy",
            "local_location_identifier": "Location",
        }
        # Define standard units
        standard_units = {
            "analyser_model": None,
            "analyser_slit_width": None,
            "analyser_slit_width_identifier": None,
            "analyser_eV": "eV",
            "analyser_step_size": "eV",
            "analyser_PE": "eV",
            "analyser_sweeps": None,
            "analyser_dwell": "ms",
            "analyser_lens_mode": None,
            "analyser_acquisition_mode": None,
            "analyser_polar": "deg",
            "analyser_tilt": "deg",
            "analyser_azi": "deg",
            "analyser_deflector_parallel": "deg",
            "analyser_deflector_perp": "deg",
            "photon_hv": "eV",
        }

        # If custom key mappings or units have been supplied, update the standard keys
        standard_keys.update(cls._SES_metadata_key_mappings)
        standard_units.update(cls._SES_metadata_units)

        # Extract metadata values and give them units where appropriate
        # - if a single key is given, extract the value directly
        # - if a list of keys is given, try and extract and return all non-None values
        # - if a callable is given, call the function with the metadata dictionary and
        # return the result
        # - add units if given and extracted value is not None
        metadata_dict = {}
        units_failure = []

        def try_parse_to_float(value):
            try:
                return float(value)
            except (ValueError, TypeError):
                return value

        # Handle extracting metadata based on arguments in the standard_keys dictionary
        for peaks_key, ses_key in standard_keys.items():
            value = None
            if isinstance(ses_key, str):
                value = try_parse_to_float(metadata_dict_SES_keys.get(ses_key))
            elif callable(ses_key):
                value = ses_key(try_parse_to_float(metadata_dict_SES_keys))
            elif isinstance(ses_key, list):
                value = [
                    try_parse_to_float(metadata_dict_SES_keys.get(key))
                    for key in ses_key
                    if metadata_dict_SES_keys.get(key) is not None
                ]

                if len(value) == 1:
                    value = value[0]
                elif len(value) == 0:
                    value = None
                else:
                    try:
                        value = np.asarray(value)
                    except (ValueError, TypeError):
                        value = str(value)

            # Try to add units if they are given and the value is not None
            if value is not None and standard_units.get(peaks_key):
                try:
                    value = value * ureg(standard_units[peaks_key])
                except TypeError:
                    units_failure.append(peaks_key)  # Keep track of errors
            metadata_dict[peaks_key] = value

        # Warn if any units failed to be added
        if units_failure:
            analysis_warning(
                f"Failed to add units to the following metadata items: {units_failure}",
                "warning",
                "Failed to add units to metadata",
            )

        return metadata_dict



"""Miscellaneous helper functions."""

import contextlib
import time
import warnings

import xarray as xr
from dask.callbacks import Callback
from IPython.display import Javascript, Markdown, display
from termcolor import colored
from tqdm.auto import tqdm


def analysis_warning(text, warn_type="info", title="Analysis info", quiet=False):
    """Tool to display a string as a warning in a formatted box.

    Parameters
    ------------
    text : str
        Text string to be displayed.

    warn_type : str, optional
        Warning box type:
            info : blue box (default)
            warning : yellow box
            success : green box
            danger : red box

    title : str, optional
        Title to be displayed in box. Defaults to `Analysis info`.

    quiet : bool, optional
        Determines whether the warning is displayed. Defaults to False.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Display a blue warning box informing the user that the data is being loaded
        pks.analysis_warning('Loading file into DataArray format.', warn_type='info', title='Loading info')

        # Display a red warning box informing the user that the fitting analysis failed
        pks.analysis_warning('Fitting result could not converge.', warn_type='danger', title='Analysis info')

    """

    # Display warning
    if not quiet:
        display(
            Markdown(
                '<div class="alert alert-block alert-%s"><b>%s: </b> %s </div>'
                % (warn_type, title, text)
            )
        )


def dequantify_quantify_wrapper(func):
    def wrapper(*args, **kwargs):
        # Apply dequantify to the first argument
        args = (args[0].pint.dequantify(),) + args[1:]

        # Call the original function
        result = func(*args, **kwargs)

        # Apply quantify to the result
        if isinstance(result, (xr.DataArray, xr.Dataset)):
            result = result.pint.quantify()
            # Check if there is a single-length coord with Units that should be dequantified
            for dim, coord in result.coords.items():
                if coord.size == 1:
                    result = result.assign_coords({dim: coord.pint.dequantify()})
        return result

    return wrapper


@contextlib.contextmanager
def silence_unit_warnings():
    from pint import UnitStrippedWarning

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UnitStrippedWarning)

        yield


def format_colored_dict(d, indent_level=0, col_cycle=0):
    """Recursive function to format dictionary with colored keys."""
    colours = ["green", "blue", "red", "yellow"]
    indent = "    " * indent_level
    current_color = colours[col_cycle % len(colours)]  # Cycle through colors
    formatted_str = ""
    for key, value in d.items():
        if isinstance(value, dict):  # Nested dictionary (recursive case)
            formatted_str += f"{indent}{colored(key, current_color)}:\n"
            formatted_str += format_colored_dict(value, indent_level + 1, col_cycle + 1)
        else:  # Base case (simple value)
            formatted_str += (
                f"{indent}{colored(key, current_color)}: {colored(value, 'black')}\n"
            )
    return formatted_str


def make_cell(text, below=True, execute=True):
    """Generate a new cell in the Jupyter notebook, directly below or above the currently running cell.

    Parameters
    ------------
    text : str
        Code to be generated in new cell.

    below : bool, optional
        Determines whether the new cell is generated below the currently running cell (as opposed to above).
        Defaults to True.

    execute : bool, optional
        Determines whether the newly generated cell is automatically executed. Defaults to True.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        # Generate a cell below which creates variable a that is equal to 2, and execute it
        make_cell('a=2')

        # Generate a cell above which calculates 5+7, but do not execute it
        make_cell('5+7', below=False, execute=False)

    """

    # Ensure text is in an appropriate format
    text = text.replace("\n", "\\n").replace("'", "\\'")

    if below:
        if execute:
            # Generate and execute new cell below
            display(
                Javascript(
                    """
            var cell = IPython.notebook.insert_cell_below('code')
            cell.set_text('"""
                    + text
                    + """')
            cell.execute()
            """
                )
            )
        else:
            # Generate but do not execute new cell below
            display(
                Javascript(
                    """
            var cell = IPython.notebook.insert_cell_below('code')
            cell.set_text('"""
                    + text
                    + """')
            """
                )
            )
    else:
        if execute:
            # Generate and execute new cell above
            display(
                Javascript(
                    """
            var cell = IPython.notebook.insert_cell_above('code')
            cell.set_text('"""
                    + text
                    + """')
            cell.execute()
            """
                )
            )
        else:
            # Generate but do not execute new cell above
            display(
                Javascript(
                    """
            var cell = IPython.notebook.insert_cell_above('code')
            cell.set_text('"""
                    + text
                    + """')
            """
                )
            )


class DaskTQDMProgressBar(Callback):
    """
    A global Dask progress bar using tqdm that mimics Dask's ProgressBar(minimum=...).
    """

    def __init__(self, desc="Dask compute", minimum=1.0, mininterval=0.1):
        """
        Parameters
        ------------
        desc : str, optional
            Label for the progress bar. Defaults to "Dask compute".
        minimum : float, optional
            Delay (in seconds) before showing the bar. Defaults to 1.0.
        mininterval : float, optional
            Minimum interval between visual updates (in seconds). Defaults to 0.1.
        """
        self.desc = desc
        self.minimum = minimum
        self.mininterval = mininterval
        self._tqdm = None
        self._start_time = None
        self._pending_updates = 0
        self._showing = False

    def _start_state(self, dsk, state):
        self.total = sum(
            len(state[k]) for k in ["ready", "waiting", "running", "finished"]
        )
        self._start_time = time.time()
        self._pending_updates = 0
        self._showing = False

    def _posttask(self, key, result, dsk, state, id):
        self._pending_updates += 1
        elapsed = time.time() - self._start_time

        if not self._showing and elapsed >= self.minimum:
            self._tqdm = tqdm(
                total=self.total,
                desc=self.desc,
                leave=True,
                mininterval=self.mininterval,
            )
            self._tqdm.update(self._pending_updates)
            self._showing = True
        elif self._showing:
            self._tqdm.update(1)

    def _finish(self, dsk, state, errored):
        if self._tqdm:
            self._tqdm.close()



import pint
import pint_xarray

ureg = pint_xarray.unit_registry


def _ensure_units(da, t, unit=None):
    """Ensure that the time is in units, taking the units of the time axis of the data by default."""
    if isinstance(t, (int, float)):
        unit = unit or da.t.units
        return t * ureg(unit)
    elif isinstance(t, pint.Quantity) and unit:
        return t.to(unit)
    return t


def mean(da):
    """Calculate the mean of the data over the time dimension."""
    mean = da.mean("t", keep_attrs=True)
    return mean.history.assign(
        f"Integrated over time from {da.t.min().data * da.t.units} to {da.t.max().data * da.t.units}"
    )


def static(da, t_static=None):
    """Calculate the static spectrum from a time-resolved experiment.
    Assumes that all data points recorded for a time < t_static are equilibrium.

    Parameters
    -----------
    da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t".

    t_static : pint.Quantity or float, optional
        time point to assume static up to. If no units are provided, the units of the data are assumed.
        default = -250 fs
    """
    if t_static is None:
        t_static = -250.0 * ureg("fs")
    t_static = _ensure_units(da, t_static)
    t_static_axis_units = t_static.to(da.t.units).magnitude
    return (
        da.sel(t=slice(None, t_static_axis_units))
        .mean("t", keep_attrs=True)
        .history.assign(f"Static ARPES calculated from data up to {t_static}")
    )


def diff(da, t_select=None, t_static=None):
    """Calculate the difference spectrum of the data at some time point or time window and the static data.

    Parameters
    -----------
    da : xarray.DataArray
        time-resolved data, with a time axis with dimension "t".

    t_select : pint.Quantity or float or slice() or None, optional
        time to select data over. If no units are provided, the units of the data are assumed.
        If a slice is provided, the mean over the slice is calculated.
        If None, the difference of the full data cube is calculated against the static data.
        default = None

    t_static : pint.Quantity or float, optional
        time point to assume static up to. If no units are provided, the units of the data are assumed.
        default = -250 fs
    """
    if t_static is None:
        t_static = -250.0 * ureg("fs")

    axis_units = da.t.units
    t_select_ = t_select
    if isinstance(t_select, slice):
        t0 = (
            t_select.start.to(axis_units).magnitude
            if isinstance(t_select.start, pint.Quantity)
            else t_select.start
        )
        t1 = (
            t_select.stop.to(axis_units).magnitude
            if isinstance(t_select.stop, pint.Quantity)
            else t_select.stop
        )
        t_select = slice(t0, t1)
    elif isinstance(t_select, pint.Quantity):
        t_select = t_select.to(axis_units).magnitude

    static_ = static(da, t_static)
    if isinstance(t_select, slice):
        excited = da.sel(t=t_select).mean("t", keep_attrs=True)
    elif t_select is not None:
        excited = da.sel(t=t_select, method="nearest")
    else:
        excited = da
    diff = excited - static_
    return diff.history.assign(
        f"Difference spectrum: {t_select_ or 'all'} - static data (t<{t_static})"
    )



import numpy as np
import pint
import pint_xarray

from peaks.core.fileIO.base_data_classes.base_manipulator_class import (
    BaseManipulatorDataLoader,
)
from peaks.core.fileIO.base_data_classes.base_photon_source_classes import (
    BasePhotonSourceDataLoader,
)
from peaks.core.fileIO.base_data_classes.base_temperature_class import (
    BaseTemperatureDataLoader,
)
from peaks.core.metadata.base_metadata_models import (
    ARPESAnalyserAnglesMetadataModel,
    ARPESAnalyserMetadataModel,
    ARPESCalibrationModel,
    ARPESDeflectorMetadataModel,
    ARPESMetadataModel,
    ARPESScanMetadataModel,
    ARPESSlitMetadataModel,
    NamedAxisMetadataModel,
)
from peaks.core.utils.misc import analysis_warning

# Define the unit registry
ureg = pint_xarray.unit_registry


class BaseARPESDataLoader(
    BasePhotonSourceDataLoader, BaseTemperatureDataLoader, BaseManipulatorDataLoader
):
    """Base class for data loaders for ARPES systems. Assume a cryo-manipulator and photon source

    Subclasses should define the `_load_analyser_metadata` method to return a dictionary of relevant metadata
    values with keys of the form in `analyser_item` where `item` is the names in the `_analyser_attributes` list,
    i.e. is given in :class:`peaks` convention. This method should return values as :class:`pint.Quantity` objects
    where possible to ensure units are appropriately captured and propagated. Alternatively, the main `_load_metadata`
    method can be overwritten to return the full metadata dictionary, including manipulator metadata.

    Subclasses should add any additional analyser attributes via the `_add_analyser_attributes` class variable,
    providing a list of additional attributes.

    Subclasses should also define the `_analyser_slit_angle` class variable if this is fixed. If custom logic is
    required, this should be left as None and the logic handled within the metadata loading.

    See Also
    --------
    BaseDataLoader
    BaseDataLoader._load_metadata
    BasePhotonSourceDataLoader
    BaseTemperatureDataLoader
    BaseManipulatorDataLoader
    """

    # Define class variables
    _loc_name = "Default ARPES"
    _dtype = "float32"
    _dorder = "C"
    # Any sign and name conventions to apply to the analyser cf. `peaks` convention
    _analyser_sign_conventions = {}
    _analyser_name_conventions = {}
    # Define analyser slit angle in subclass, otherwise it must be passed to the metadata dict from the metadata loader
    _analyser_slit_angle = None
    # Core attributes
    _analyser_attributes = [
        "model",
        "slit_width",
        "slit_width_identifier",
        "eV",
        "step_size",
        "PE",
        "sweeps",
        "dwell",
        "lens_mode",
        "acquisition_mode",
        "eV_type",
        "scan_command",
        "polar",
        "tilt",
        "azi",
        "deflector_parallel",
        "deflector_perp",
    ]
    _desired_dim_order = [
        "scan_no",
        "hv",
        "temperature_sample",
        "temperature_cryostat",
        "x3",
        "x2",
        "x1",
        "polar",
        "tilt",
        "azi",
        "y_scale",
        "deflector_perp",
        "eV",
        "deflector_parallel",
        "theta_par",
    ]
    _default_dimensions = {
        "hv": "eV",
        "temperature_sample": "K",
        "temperature_cryostat": "K",
        "x3": "mm",
        "x2": "mm",
        "x1": "mm",
        "polar": "deg",
        "tilt": "deg",
        "azi": "deg",
        "y_scale": "mm",
        "deflector_perp": "deg",
        "eV": "eV",
        "deflector_parallel": "deg",
        "theta_par": "deg",
    }
    _analyser_include_in_metadata_warn = [
        "eV",
        "step_size",
        "PE",
        "sweeps",
        "dwell",
        "azi",
    ]  # List of analyser attributes to include in metadata warning
    _metadata_parsers = [
        "_parse_analyser_metadata",
        "_parse_manipulator_metadata",
        "_parse_photon_metadata",
        "_parse_temperature_metadata",
    ]  # List of metadata parsers to apply

    @property
    def analyser_attributes(self):
        return self._analyser_attributes

    @classmethod
    def _load(cls, fpath, lazy, metadata, quiet, **kwargs):
        # Raise an exception if the manipulator doesn't have at least the core 6 axes required in other data handling
        required_axis_list = ["polar", "tilt", "azi", "x1", "x2", "x3"]
        if not hasattr(cls, "_manipulator_axes") or not all(
            item in cls._manipulator_axes for item in required_axis_list
        ):
            raise ValueError(
                f"ARPES data loaders are expected to subclass a BaseManipulatorDataLoader and include all of "
                f"{required_axis_list} in the `_manipulator_axes` class variable. If the physical axis is missing, "
                f"this can be represented by not passing a name for that axis in the `_manipulator_name_conventions` "
                f"class variable, but the axis is required in `_manipulator_axes` to be able to define normal "
                f"emissions for that axis, required for e.g. ARPES k-conversions."
            )

        # Run the main _load method returning a DataArray
        da = super()._load(fpath, lazy, metadata, quiet, **kwargs)

        # Try to parse to counts/s if possible
        try:
            if "count" in str(da.pint.units) and "/s" not in str(da.pint.units):
                t = (
                    da.metadata.analyser.scan.dwell.to("s")
                    * da.metadata.analyser.scan.sweeps
                )
                da /= t
        except (ValueError, AttributeError, TypeError):
            pass

        # Set default on dimensions if not already set
        for dim in da.dims:
            if not hasattr(da[dim], "units") and dim in cls._default_dimensions:
                da = da.pint.quantify({dim: cls._default_dimensions[dim]})

        return da

    @classmethod
    def _parse_analyser_metadata(cls, metadata_dict):
        """Parse metadata specific to the analyser."""

        # Get the analyser slit angle from default option for loader if specified and not already in metadata dict
        if not metadata_dict.get("analyser_azi"):
            metadata_dict["analyser_azi"] = cls._analyser_slit_angle

        # Make and populate the analyser metadata model
        arpes_metadata = ARPESMetadataModel(
            analyser=ARPESAnalyserMetadataModel(
                model=metadata_dict.get("analyser_model"),
                slit=ARPESSlitMetadataModel(
                    width=metadata_dict.get("analyser_slit_width"),
                    identifier=metadata_dict.get("analyser_slit_width_identifier"),
                ),
            ),
            scan=ARPESScanMetadataModel(
                eV=metadata_dict.get("analyser_eV"),
                step_size=metadata_dict.get("analyser_step_size"),
                PE=metadata_dict.get("analyser_PE"),
                sweeps=metadata_dict.get("analyser_sweeps"),
                dwell=metadata_dict.get("analyser_dwell"),
                lens_mode=metadata_dict.get("analyser_lens_mode"),
                acquisition_mode=metadata_dict.get("analyser_acquisition_mode"),
                eV_type=metadata_dict.get("analyser_eV_type"),
            ),
            angles=ARPESAnalyserAnglesMetadataModel(
                polar=metadata_dict.get("analyser_polar"),
                tilt=metadata_dict.get("analyser_tilt"),
                azi=metadata_dict.get("analyser_azi"),
            ),
            deflector=ARPESDeflectorMetadataModel(
                parallel=NamedAxisMetadataModel(
                    value=metadata_dict.get("analyser_deflector_parallel"),
                    local_name=cls._analyser_name_conventions.get("deflector_parallel"),
                ),
                perp=NamedAxisMetadataModel(
                    value=metadata_dict.get("analyser_deflector_perp"),
                    local_name=cls._analyser_name_conventions.get("deflector_perp"),
                ),
            ),
        )

        # Return the model, and a list of any metadata that should be warned if missing
        return {
            "_calibration": ARPESCalibrationModel(),
            "_analyser": arpes_metadata,
        }, [f"analyser_{i}" for i in cls._analyser_include_in_metadata_warn]

    # Methods to parse manipulator reference values (normal emissions etc.)
    @classmethod
    def _get_sign_convention(cls, axis_key):
        """
        Retrieve the sign convention for a given axis.
        """
        if "deflector" in axis_key or "theta_par" in axis_key or "ana_" in axis_key:
            return cls._analyser_sign_conventions.get(axis_key, 1)
        else:
            return cls._manipulator_sign_conventions.get(axis_key, 1)

    @classmethod
    def _group_axes(cls, da):
        """
        Group axes based on the slit orientation.
        """
        if da.metadata.analyser.angles.azi == 0:
            # Slit angle 0, along the polar axis
            polar_group = {"polar", "deflector_perp", "ana_polar"}
            tilt_group = {"tilt", "deflector_parallel", "theta_par", "ana_tilt"}
            azi_group = {"azi", "azi_offset"}
        else:
            # Slit angle 90, perpendicular to the polar axis
            polar_group = {"polar", "deflector_parallel", "theta_par", "ana_polar"}
            tilt_group = {"tilt", "deflector_perp", "ana_tilt"}
            azi_group = {"azi", "azi_offset"}
        return {"polar": polar_group, "tilt": tilt_group, "azi": azi_group}

    @classmethod
    def _get_relative_sign_conventions(cls, da, axis_key, axis_group):
        """
        Get the relative sign convention for a given axis to its physical manipulator axis
        axis_group is either an axis group or a string
        """
        if isinstance(axis_group, str):
            # Handle analyser angle to manipulator angle sign conventions used in get
            # Ishida and Shin convetions function
            manipulator_axis = axis_group
        elif isinstance(axis_group, set):
            manipulator_axis = next(
                (key for key in ["polar", "tilt", "azi"] if key in axis_group), None
            )
            axis_groups = cls._group_axes(da)
            polar_group = axis_groups.get("polar")

        manipulator_axis_sign_convention = cls._get_sign_convention(manipulator_axis)
        axis_key_sign_convention = cls._get_sign_convention(axis_key)
        sign_product = manipulator_axis_sign_convention * axis_key_sign_convention

        if isinstance(axis_group, str):
            if axis_key == f"ana_{manipulator_axis}":
                return -1 if sign_product == 1 else 1

        if axis_key == manipulator_axis:
            return 1

        if manipulator_axis == "azi" and axis_key == "azi_offset":
            return 1  # always 1 as already corrected in disp 3d panel

        if "polar" in polar_group and "deflector_perp" in polar_group:
            # Analyser type must be either I or Ip
            if manipulator_axis == "polar":
                if axis_key == "ana_polar":
                    # Type I
                    return -1 if sign_product == 1 else 1
                elif axis_key == "deflector_perp":
                    # Type Ip
                    return 1 if sign_product == 1 else -1
            elif manipulator_axis == "tilt":
                if axis_key == "theta_par":
                    # Same for both I and Ip
                    return -1 if sign_product == 1 else 1
                elif axis_key == "deflector_parallel":
                    return -1  # consistent with theta_par
                elif axis_key == "ana_tilt":
                    return -1 if sign_product == 1 else 1
        else:
            # Analyser type must be either II or IIp
            if manipulator_axis == "tilt":
                if axis_key == "ana_tilt":
                    # Type II
                    return -1 if sign_product == 1 else 1
                elif axis_key == "deflector_perp":
                    # Type IIp
                    return -1 if sign_product == 1 else 1
            elif manipulator_axis == "polar":
                if axis_key == "theta_par":
                    return -1 if sign_product == 1 else 1
                elif axis_key == "deflector_parallel":
                    return -1  # # consistent with theta_par
                elif axis_key == "ana_polar":
                    return -1 if sign_product == 1 else 1

        raise ValueError(
            f"Unexpected axis combination: manipulator_axis is {manipulator_axis}, axis_key is {axis_key}"
        )

    @classmethod
    def _parse_reference_value(cls, da, axis_key, value, axis_group):
        """
        Parse the manipulator reference value for a given axis.
        """

        if axis_key in ["polar", "tilt", "azi"] and axis_key not in da.dims:
            # The axis is a core manipulator axis and is not a scannable, so this should be the actual normal emission
            return value

        reference_angles = []
        reference_signs = []

        # Primary axis value (user-specified)
        # Read the offsets from the GUI and assign relative sign conventions
        reference_angles.append(value)
        reference_signs.append(
            cls._get_relative_sign_conventions(da, axis_key, axis_group)
        )

        # Secondary axes (from metadata)
        for axis in axis_group:
            if axis == axis_key:
                continue  # Skip the primary axis
            angle_value = cls._get_axis_value(da, axis)
            if angle_value is not None:
                reference_angles.append(angle_value)
                reference_signs.append(
                    cls._get_relative_sign_conventions(da, axis, axis_group)
                )

        # Convert angles to consistent units and sum
        total_reference_angle = cls._sum_angles(reference_angles, reference_signs)
        return total_reference_angle

    @classmethod
    def _get_axis_value(cls, da, axis):
        """
        Retrieve the value of an axis from the data array or metadata.
        """
        if hasattr(da, axis):
            data = getattr(da, axis).pint.to("deg")
            return (
                data.data
                if isinstance(data.data, pint.Quantity)
                else data.data * ureg("deg")
            )

        else:
            # Attempt to get from metadata
            if "deflector" in axis:
                return getattr(da.metadata.analyser.deflector, axis.split("_")[1]).value
            elif "ana_" in axis:
                return getattr(da.metadata.analyser.angles, axis.split("_")[1])
            elif hasattr(da.metadata.manipulator, axis):
                return getattr(da.metadata.manipulator, axis).value
            else:
                return None  # Axis value not found

    @classmethod
    def _sum_angles(cls, angles, signs):
        """
        Sum angles with appropriate sign conventions.
        """
        total_angle = 0.0 * ureg("deg")
        for angle, sign in zip(angles, signs, strict=True):
            try:
                angle_value = angle.to("deg")
            except AttributeError:
                angle_value = angle * ureg("deg")
            total_angle += angle_value * sign
        return total_angle

    @classmethod
    def _parse_manipulator_references(cls, da, specified_values):
        """
        Parse manipulator reference values based on specified parameters.
        """
        axis_groups = cls._group_axes(da)
        reference_values = {}

        for axis, group in axis_groups.items():
            # Check if any axis in the group is specified
            specified_axis = None
            for ax in group:
                if ax in specified_values:
                    specified_axis = ax
                    break
            if specified_axis:
                if (
                    specified_axis in ["polar", "tilt", "azi"]
                    and specified_axis not in da.dims
                ):
                    # If the axis is a core manipulator axis and not a scannable,
                    # use the specified value directly
                    final_value = specified_values[specified_axis]
                else:
                    value = specified_values[specified_axis]
                    # Parse the reference value
                    total_reference = cls._parse_reference_value(
                        da, specified_axis, value, group
                    )

                    final_value = total_reference
                try:
                    reference_values[axis] = final_value.to("deg")
                except AttributeError:
                    reference_values[axis] = final_value * ureg("deg")
        return reference_values

    @classmethod
    def _get_angles_Ishida_Shin(cls, da, quiet=False):
        """
        Convert angles to the conventions of Ishida and Shin.
        """

        missing_values = {}
        missing_references = {}

        # Helper functions
        def get_angle(da, axis, default=None, add_to_warning_list=True):
            """Get the angle value for a given axis.

            If the value is missing, return the default value

            Parameters
            ----------
            da : xarray.DataArray
                The data array.
            axis : str
                The axis name.
            default : pint.Quantity, optional
                The default value to return if the axis value is missing.

            Returns
            -------
            pint.Quantity
                The angle value.
            """
            if default is None:
                default = 0.0 * ureg("rad")
            value = cls._get_axis_value(da, axis)
            if value is None:
                if add_to_warning_list:
                    missing_values[axis] = str(default)
                return default
            else:
                return value.to("rad") * cls._get_sign_convention(axis)

        def get_reference_angle(da, axis, default=None):
            """Get the reference angle value for a given axis.

            If the value is missing, return the default value.

            Parameters
            ----------
            da : xarray.DataArray
                The data array.
            axis : str
                The axis name.
            default : pint.Quantity, optional
                The default value to return if the axis value is missing.

            Returns
            -------
            pint.Quantity
                The angle value.

            """
            if default is None:
                default = 0.0 * ureg("rad")

            value = getattr(da.metadata.manipulator, axis).reference_value
            if value is None:
                missing_references[axis] = str(default)
                return default
            else:
                return value.to("rad") * cls._get_sign_convention(axis)

        # Retrieve manipulator angles
        manipulator_angles = {
            axis: get_angle(da, axis) for axis in ["polar", "tilt", "azi"]
        }

        # Retrieve analyser angles
        analyser_angles = {
            axis: get_angle(da, f"ana_{axis}", add_to_warning_list=False)
            for axis in ["polar", "tilt", "azi"]
        }

        # Retrieve deflector angles
        deflector_angles = {
            axis: get_angle(da, f"deflector_{axis}", add_to_warning_list=False)
            for axis in ["parallel", "perp"]
        }

        # Determine the type based on analyser azi angle
        analyser_azi = analyser_angles.get("azi", 0)
        if analyser_azi == 0:
            if any(
                value is not None
                and isinstance(value, pint.Quantity)
                and np.any(value.magnitude != 0)
                for value in deflector_angles.values()
            ):
                type_ = "Ip"
                alpha = get_angle(da, "theta_par") + deflector_angles["parallel"]
                beta = deflector_angles["perp"]
                chi = manipulator_angles["polar"] + analyser_angles[
                    "polar"
                ] * cls._get_relative_sign_conventions(da, "ana_polar", "polar")
                xi = manipulator_angles["tilt"] + analyser_angles[
                    "tilt"
                ] * cls._get_relative_sign_conventions(da, "ana_tilt", "tilt")
                beta_0 = None
                chi_0 = get_reference_angle(da, "polar")
                xi_0 = get_reference_angle(da, "tilt")
            else:
                type_ = "I"
                alpha = get_angle(da, "theta_par")
                beta = manipulator_angles["polar"] + analyser_angles[
                    "polar"
                ] * cls._get_relative_sign_conventions(da, "ana_polar", "polar")
                chi = None
                xi = manipulator_angles["tilt"] + analyser_angles[
                    "tilt"
                ] * cls._get_relative_sign_conventions(da, "ana_tilt", "tilt")
                beta_0 = get_reference_angle(da, "polar")
                chi_0 = None
                xi_0 = get_reference_angle(da, "tilt")
        else:
            if any(
                value is not None
                and isinstance(value, pint.Quantity)
                and np.any(value.magnitude != 0)
                for value in deflector_angles.values()
            ):
                type_ = "IIp"
                alpha = get_angle(da, "theta_par") + deflector_angles["parallel"]
                beta = deflector_angles["perp"]
                chi = manipulator_angles["polar"] + analyser_angles[
                    "polar"
                ] * cls._get_relative_sign_conventions(da, "ana_polar", "polar")
                xi = manipulator_angles["tilt"] + analyser_angles[
                    "tilt"
                ] * cls._get_relative_sign_conventions(da, "ana_tilt", "tilt")
                beta_0 = None
                chi_0 = get_reference_angle(da, "polar")
                xi_0 = get_reference_angle(da, "tilt")
            else:
                type_ = "II"
                alpha = get_angle(da, "theta_par")
                beta = manipulator_angles["tilt"] + analyser_angles[
                    "tilt"
                ] * cls._get_relative_sign_conventions(da, "ana_tilt", "tilt")
                chi = None
                xi = manipulator_angles["polar"] + analyser_angles[
                    "polar"
                ] * cls._get_relative_sign_conventions(da, "ana_polar", "polar")
                beta_0 = get_reference_angle(da, "tilt")
                chi_0 = None
                xi_0 = get_reference_angle(da, "polar")

        delta = manipulator_angles["azi"]
        delta_0 = get_reference_angle(da, "azi", default=delta)

        # Check and warn for missing values
        if not quiet and missing_values:
            analysis_warning(
                missing_values,
                "warning",
                "Missing manipulator angle values. Using default values.",
            )
        if not quiet and missing_references:
            analysis_warning(
                missing_references,
                "warning",
                "Missing manipulator reference values. Using default values.",
            )

        angle_dict = {
            "type": type_,
            "alpha": alpha,
            "beta": beta,
            "beta_0": beta_0,
            "chi": chi,
            "chi_0": chi_0,
            "xi": xi,
            "xi_0": xi_0,
            "delta": delta,
            "delta_0": delta_0,
        }

        return angle_dict



import importlib
import sys
from functools import wraps

import xarray as xr

from peaks.core.utils.misc import analysis_warning


class LazyAccessorDescriptor:
    """Descriptor for lazy-loading accessors, providing documentation and lazy execution."""

    def __init__(self, func_name, module_name, accessor_class):
        self.func_name = func_name
        self.module_name = module_name
        self.accessor_class = accessor_class
        self._func = None  # Cached function after loading

    def __get__(self, instance, owner):
        if instance is None:
            return self  # Accessed on the class, return the descriptor itself
        return self.accessor_class(instance, self.module_name, self.func_name)

    @property
    def __doc__(self):
        """Provide the docstring of the target function."""
        if self._func is None:
            self._load_function()
        return self._func.__doc__

    def _load_function(self):
        """Load the function and cache it."""
        module = importlib.import_module(self.module_name)
        self._func = getattr(module, self.func_name)

    def __dir__(self):
        """Provide autocompletion for the target function's attributes."""
        if self._func is None:
            self._load_function()
        return dir(self._func)

    def __getattr__(self, name):
        """Forward attribute access to the target function."""
        if self._func is None:
            self._load_function()
        return getattr(self._func, name)


class PeaksDirectCallAccessor:
    """Lazy-loading accessor for functions applied to xarray DataArrays."""

    def __init__(self, data_array, module_name, func_name):
        self.data_array = data_array
        self.module_name = module_name
        self.func_name = func_name
        self._func = None  # Cache for the function

    def _load_function(self):
        """Loads the function from the specified module, updating the cache if necessary."""
        # Always reload the module to ensure we get the latest version (compatible with %autoreload)
        module = importlib.import_module(self.module_name)
        if module.__name__ in sys.modules:
            importlib.reload(sys.modules[module.__name__])
        self._func = getattr(module, self.func_name)
        # Update the __doc__ attribute
        self.__doc__ = (
            ":::{note} "
            f"This is an accessor to the function `{self.module_name}.{self.func_name}` acting "
            f"directly on the xarray object. (Note, the :class:`xarray.DataArray`, :class:`xarray.Dataset` "
            f"or :class:`xarray.DataTree` does not now need to be explicitly passed to the function)."
            f":::\n\n"
            f"{self._func.__doc__ or ''}"
        )

    @property
    def func(self):
        """Property to access the loaded function, loading it if necessary."""
        if self._func is None:
            self._load_function()
        return self._func

    def __call__(self, *args, **kwargs):
        """Calls the loaded function with the DataArray and provided arguments."""
        return self.func(self.data_array, *args, **kwargs)

    def __dir__(self):
        """Provide autocompletion for the original function's attributes."""
        return dir(self.func)

    def __getattr__(self, name):
        """Forward attribute access to the loaded function if available."""
        return getattr(self.func, name)


class PeaksDataTreeIteratorAccessor(PeaksDirectCallAccessor):
    """Accessor that applies the function to each node in a DataTree.

    Assumes:
    - A hollow tree structure.
    - Leaves hold `xarray.Dataset` objects.
    - If the Dataset is the primary data container, it will have metadata attributes.
    - Otherwise, the Dataset will hold a single `xarray.DataArray` with the data variable `data`.
    """

    def __init__(self, data_tree, module_name, func_name):
        super().__init__(data_tree, module_name, func_name)
        self.data_tree = data_tree  # Overwrite data_array with data_tree

    def __call__(self, *args, **kwargs):
        """Iterates over DataTree nodes and applies the function."""
        func = self.func  # Ensure the function is loaded

        def apply_func(node):
            if hasattr(node, "_scan"):
                # Apply the function to the node directly
                return func(node, *args, **kwargs)
            elif hasattr(node, "data") and len(node) == 1:
                # Map over the Dataset which contains only a single DataArray
                return node.map(lambda da: func(da, *args, **kwargs))
            elif len(node) == 0:
                pass
            else:
                raise ValueError(
                    "For use of this `peaks` accessor, leaf nodes of the xr.DataTree must be a xr.Dataset with "
                    "a single xr.DataArray with key `data`, or the Dataset itself must be the primary data container, "
                    f"itself holding metadata. The current node does not meet these requirements: {node}."
                )

        # Use DataTree's map_over_datasets method to apply the function
        return self.data_tree.map_over_datasets(apply_func)

    def __dir__(self):
        """Provide autocompletion for the original function's attributes."""
        return dir(self.func)

    def __getattr__(self, name):
        """Forward attribute access to the loaded function if available."""
        return getattr(self.func, name)


def _pass_function_to_xarray_class_accessor(func_name, module_name):
    # Import the module and get the function once
    module = importlib.import_module(module_name)
    func = getattr(module, func_name)

    @wraps(func)
    def method(self, *args, **kwargs):
        # Check if the object is a DataTree and use PeaksDataTreeIteratorAccessor
        if isinstance(self._obj, xr.DataTree):
            accessor = PeaksDataTreeIteratorAccessor(self._obj, module_name, func_name)
            return accessor(*args, **kwargs)
        else:
            # Call the original function with self._obj
            return func(self._obj, *args, **kwargs)

    # Modify the docstring to include the custom note
    note = (
        ":::{note} "
        f"This is an accessor to the function `{module_name}.{func_name}` acting "
        f"directly on the :class:`xarray` object. (Note, the :class:`xarray.DataArray`, :class:`xarray.Dataset` "
        f"or :class:`xarray.DataTree` does not now need to be explicitly passed to the function)."
        f":::\n\n"
    )
    method.__doc__ = note + (func.__doc__ or "")
    return method


def register_lazy_accessor(
    func_name, module_name, cls, accessor_class=PeaksDirectCallAccessor
):
    """Register a lazy-loading accessor as a descriptor on the class."""

    # Check if the accessor already exists on cls and issue a warning if so
    if hasattr(cls, func_name):
        analysis_warning(
            f"Registration of accessor under name '{func_name}' for type "
            f"{cls.__module__}.{cls.__name__} is overriding an existing attribute with the same name.",
            title="Registration of accessor",
            warn_type="warning",
        )

    # Register the accessor descriptor on cls
    setattr(
        cls, func_name, LazyAccessorDescriptor(func_name, module_name, accessor_class)
    )


def register_accessor(cls):
    """Decorator (function wrapper) used to allow a function to be used as an object-oriented programming method.

    Parameters
    ------------
    cls : object
        Here, this will be an:class:`xarray.DataArray` or similar. This allows a function to be applied to that
        class as a method.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks
        from peaks.core.accessors.accessor_methods import register_accessor

        my_data = pks.load('my_file.ibw')

        # Turn the function (data_plus_1) into a xarray.DataArray method
        @register_accessor(xr.DataArray)
        def data_plus_1(data):
            return data+1

        # Create a new xarray equal to my_data + 1
        my_data_plus_1 = my_data.data_plus_1()

    """

    # decorator for the function - func
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        # Give a warning if already exists
        if hasattr(cls, func.__name__):
            analysis_warning(
                f"Registration of accessor for `peaks` under name {func.__name__!r} for type "
                f"{cls.__module__}.{cls.__name__} is overriding a preexisting attribute with the same name.",
                title=" Registration of accessor",
                warn_type="warning",
            )

        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but instead the wrapper which accepts self but does exactly the same as func
        return func  # returning func means func can still also be used normally (i.e. not as a method)

    return decorator



import numbers

import numpy as np
import xarray as xr

from peaks.core.options import LocOpts
from peaks.core.utils.misc import analysis_warning

# Constants for angle names
ANGLE_NAMES = [
    "theta_par",
    "polar",
    "tilt",
    "azi",
    "ana_polar",
    "defl_par",
    "defl_perp",
]

NORM_ANGLE_NAMES = [
    "norm_polar",
    "norm_tilt",
    "norm_azi",
]

ALLOWED_ANGLES = ANGLE_NAMES + NORM_ANGLE_NAMES


def _get_conventions(data):
    """Get the angle conventions for the data.

    Parameters
    ----------
    data : xarray.DataArray
        Data for which to get the angle conventions.

    Returns
    -------
    class
        Relevant data convention Class instance.
    """

    loc = data.attrs.get("beamline")
    return LocOpts.get_conventions(loc)


def _get_raw_angles(
    data,
    angle_names,
    default_value=0,
    quiet=False,
    warn_prefix="",
):
    """
    Get the raw angles from the data, with sign conventions of the beamline.

    Parameters
    ----------
    data : xarray.DataArray
        Data from which to get the angles.
    angle_names : list
        List of angle names to retrieve from the data.
    default_value : int or float, optional
        Default value to use if an angle is missing. Defaults to 0.
    quiet : bool, optional
        Whether to suppress warnings of missing angles. Defaults to False.
    warn_prefix : str, optional
        Prefix to add to warning messages. Defaults to an empty string.

    Returns
    -------
    dict
        Dictionary of angles retrieved from the data.
    """
    angles = {}
    warn_str = ""
    for angle_name in angle_names:
        if angle_name in data.coords:
            angles[angle_name] = data.coords[angle_name].data
        else:
            angle = data.attrs.get(angle_name)
            if angle is not None:
                angles[angle_name] = angle
            else:
                angles[angle_name] = default_value
                if not quiet:
                    warn_str += f"{angle_name}: {default_value}, "
    if warn_str:
        warn_message = f"{warn_prefix}Assuming default values for missing angles: {warn_str.rstrip(', ')}."
        analysis_warning(warn_message, "warning", "Analysis warning", quiet)
    return angles


def _parse_normal_angle(angles, norm_angles, norm_angle_list, conventions):
    """
    Parse a single normal emission angle from the supplied angle specification with sign conventions of Ishida & Shin.

    Parameters
    ----------
    angles : dict
        Dictionary of angles from the data.
    norm_angles : dict
        Dictionary of normal emission angles to set.
    norm_angle_list : list
        List of angle names to consider for normal emission.
    conventions : dict
        Dictionary of angle conventions.

    Returns
    -------
    float or None
        Parsed normal emission angle (in `peaks` [Ishida&Shin compatible] co-ordinate system),
        or `None` if nothing relevant is specified for this axis.
    """
    specified_angle = [i for i in norm_angle_list if i in norm_angles]
    if len(specified_angle) > 1:
        raise ValueError(f"Only one of {norm_angle_list} can be specified.")
    elif len(specified_angle) == 0:
        return None
    angle_name = specified_angle[0]
    norm_angle = norm_angles[angle_name] * conventions.get(angle_name, 1)

    # Handle the sign conventions to make consistent with Ishida and Shin definitions
    if angle_name in ["theta_par", "azi"]:
        norm_angle *= -1
    if angle_name == "defl_perp" and conventions.get("ana_type") in ["II", "IIp"]:
        norm_angle *= -1

    # Deal with the other relevant angles
    other_angles = [i for i in norm_angle_list if i != angle_name]
    for i in other_angles:
        angle_to_add = angles.get(i, 0)
        angle_to_add *= conventions.get(i, 1)
        if isinstance(angle_to_add, numbers.Number):
            norm_angle += angle_to_add
        else:
            raise ValueError(
                f"Angle {i} is not a number. Cannot parse a unique normal emission value from the supplied data "
                f"and normal emission angle choice."
            )
    return norm_angle


def _parse_norm_angles(data, norm_angles, quiet):
    """Parse the normal emission angles from the data, given a set of angle specifiers.

    Parameters
    ----------
    data : xarray.DataArray
        Data for which to set the normal emission angles.
    norm_angles : dict
        Dictionary of angles to set normal emission from. Supplied in common peaks angle name
        (e.g. `theta_par`, `polar`, etc.).
    quiet : bool
        Whether to suppress warnings.

    Returns
    -------
    dict :
        Dictionary of normal emission angles that could be parsed from the supplied arguments.
    """

    # Get angle conventions and analyser type
    conventions = _get_conventions(data)
    ana_type = conventions.get("ana_type")

    # Check if the angles are in the correct format
    if not norm_angles:
        raise ValueError(
            "No angles supplied. Either supply a dictionary of angles or pass as keyword arguments."
        )
    invalid_angles = [i for i in norm_angles.keys() if i not in ALLOWED_ANGLES]
    if invalid_angles:
        raise ValueError(
            f"Angles {invalid_angles} are not allowed. Allowed angles are {ALLOWED_ANGLES}."
        )

    # Get the raw angles from the data
    angles = _get_raw_angles(data, ANGLE_NAMES, quiet=quiet)

    # Calculate the normal emission angles
    if "norm_polar" in norm_angles:
        norm_polar = norm_angles["norm_polar"]
    else:
        if ana_type in ["I", "Ip"]:
            norm_polar = _parse_normal_angle(
                angles, norm_angles, ["polar", "ana_polar", "defl_perp"], conventions
            )
        elif ana_type in ["II", "IIp"]:
            norm_polar = _parse_normal_angle(
                angles,
                norm_angles,
                ["polar", "ana_polar", "theta_par", "defl_par"],
                conventions,
            )
    if "norm_tilt" in norm_angles:
        norm_tilt = norm_angles["norm_tilt"]
    else:
        if ana_type in ["I", "Ip"]:
            norm_tilt = _parse_normal_angle(
                angles, norm_angles, ["theta_par", "tilt", "defl_par"], conventions
            )
        elif ana_type in ["II", "IIp"]:
            norm_tilt = _parse_normal_angle(
                angles, norm_angles, ["tilt", "defl_perp"], conventions
            )
    if "norm_azi" in norm_angles:
        norm_azi = norm_angles["norm_azi"]
    else:
        norm_azi = _parse_normal_angle(angles, norm_angles, ["azi"], conventions)

    # Make a dictionary of available normal emissions to return
    norm_emissions = {
        key: val
        for key, val in zip(
            ["norm_polar", "norm_tilt", "norm_azi"],
            [norm_polar, norm_tilt, norm_azi],
            strict=True,
        )
        if val is not None
    }

    return norm_emissions


def _set_norm_angles(
    data, norm_angles=None, quiet=False, update_in_place=True, **kwargs
):
    """
    Set the normal emission angles for the data. This can be called with updating the attributes of the existing `data`
    :class:`xarray.DataArray` (accessed via the `.norm_angles.set()` accessor) or returning a copy with the updated
    metadata (accessed via the `norm_angles.assign()`), for chaining method calls together.

    Parameters
    ----------
    data : xarray.DataArray
        Data for which to set the normal emission angles.
    norm_angles : dict, optional
        Dictionary of angles to set normal emission from. Supplied as original angle name
        (e.g. `{theta_par: 12, polar: 5}`), or using `norm_polar`, `norm_tilt`, `norm_azi`.
        Defaults to None, in which case angles must be defined as keyword arguments.
        If supplied, takes precedence over kwargs.
    quiet : bool, optional
        Whether to suppress warnings. Defaults to False.
    update_in_place : bool, optional
        Whether to update the data in place. Defaults to True.
    **kwargs : dict
        Keyword arguments of angles to set (e.g. `theta_par=12`, `polar=5`, `norm_azi=-5`).

    Returns
    -------
    xarray.DataArray : optional
        If `update_in_place` is False, returns a copy of the original `data` :class:`xarray.DataArray` with the normal
        emission angles applied. Otherwise, returns None with the emission angle metadata updated in place on the
        original :class:`xarray.DataArray`.

    Examples
    --------
    Example usage is as follows::
        import peaks as pks

        # Load data
        FS = pks.load("path/to/data")

        # Set the normal emission angles in FS
        FS.norm_angles.set(theta_par=12, polar=5)

        # Alternatively, to return a copy of the data with the normal emission angles applied
        FS_with_norm = FS.norm_angles.assign(theta_par=12, polar=5)
    """
    if norm_angles is None:
        norm_angles = kwargs

    # Parse the angles
    norm_angles_out = _parse_norm_angles(data, norm_angles, quiet)

    # History to update
    hist = (
        f"Normal emission angles set to: {norm_angles_out} from user input {norm_angles}"
    )

    if update_in_place:
        # Update the data attributes and history
        data.attrs.update(norm_angles_out)
        data.history.add(hist)
    else:
        # Assign the data attributes and history
        return data.assign_attrs(norm_angles_out).history.assign(
            hist,
        )


def _parse_manip_angles_from_norm(data, norm_angles):
    """Parse the manipulator angles from the normal emission angles.

    Parameters
    ----------
    data : xarray.DataArray
        Data for which to parse the manipulator angles.
    norm_angles : dict
        Dictionary of normal emission angles.

    Returns
    -------
    dict
        Dictionary of manipulator angles in the format {f"norm_{axis}_manip": {axis_name: norm_value}} for available
        axes.
    """
    # Get angle conventions and analyser type
    conventions = _get_conventions(data)

    # Calculate the manipulator angles from the normal emission angles
    manip_normals = dict()
    norm_polar_name = conventions.get("polar_name")
    if norm_polar_name and norm_angles.get("norm_polar") is not None:
        norm_polar_manip = {
            norm_polar_name: norm_angles["norm_polar"] * conventions.get("polar", 1)
        }
        manip_normals["norm_polar_manip"] = norm_polar_manip
    norm_tilt_name = conventions.get("tilt_name")
    if norm_tilt_name and norm_angles.get("norm_tilt") is not None:
        norm_tilt_manip = {
            norm_tilt_name: norm_angles["norm_tilt"] * conventions.get("tilt", 1)
        }
        manip_normals["norm_tilt_manip"] = norm_tilt_manip
    norm_azi_name = conventions.get("azi_name")
    if norm_azi_name and norm_angles.get("norm_azi") is not None:
        norm_azi_manip = {
            norm_azi_name: norm_angles["norm_azi"] * conventions.get("azi", 1)
        }
        manip_normals["norm_azi_manip"] = norm_azi_manip

    return manip_normals


def _get_norm_angles(data):
    """Get the normal emission angles from the data.

    Parameters
    ----------
    data : xarray.DataArray
        Data for which to get the normal emission angles.

    Returns
    -------
    dict
        Dictionary of normal emission angles in both `peaks` format and where applicable the true manipulator angles.
    """

    # Get the peaks normal emission angles from the attributes
    norm_angles = {name: data.attrs.get(name) for name in NORM_ANGLE_NAMES}

    # Parse the manipulator angles from the normal emission angles
    norm_angles.update(_parse_manip_angles_from_norm(data, norm_angles))

    return norm_angles


@xr.register_dataarray_accessor("norm_angles")
class NormAngles:
    """Custom `peaks` accessor for the normal emission metadata of a `peaks`-compatible :class:`xarray.DataArray`.

    Access via the `.norm_angles` method.

    Methods
    -------
    __call__
        Displays the currently set normal emission values of the :class:`xarray.DataArray`, including displaying the
        relevant axis notation for the physical manipulator angles.
    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __call__(self):
        """
        Displays the currently set normal emission values of self, including displaying the relevant axis notation
        for the physical manipulator angles.

        Returns
        -------
        None :
            Displays the normal emission values of the :class:`xarray.DataArray` in the console.

        Examples
        --------
        Example usage is as follows::
            import peaks as pks

            # Load data and set its normal emission
            disp = load('disp.ibw')
            disp.norm_angles.set(theta_par=12, polar=5)

            # Display the current normal emission metadata
            disp.norm_angles()

        """

        norm_angles = _get_norm_angles(self._obj)
        display_str = "<br>"
        axes = ["norm_polar", "norm_tilt", "norm_azi"]
        for axis in axes:
            if norm_angles.get(axis) is not None:
                display_str += f"&nbsp;&nbsp;&nbsp;&nbsp;{axis} = {norm_angles[axis]}"
                if f"{axis}_manip" in norm_angles:
                    manip_name = list(norm_angles[f"{axis}_manip"].keys())[0]
                    manip_value = list(norm_angles[f"{axis}_manip"].values())[0]
                    display_str += f" ({manip_name} = {manip_value})"
                display_str += "<br>"

        analysis_warning(display_str, title="Normal emission angles")

    def set(self, norm_angles=None, quiet=False, **kwargs):
        """
        Set the normal emission angles for the data, updating the data in place.

        See Also
        --------
        peaks.utils.angles._set_norm_angles

        """
        return _set_norm_angles(
            self._obj,
            norm_angles=norm_angles,
            quiet=quiet,
            update_in_place=True,
            **kwargs,
        )

    def assign(self, norm_angles=None, quiet=False, **kwargs):
        """
        Set the normal emission angles for the data, returning a copy with the updated metadata.

        See Also
        --------
        peaks.utils.angles._set_norm_angles

        """
        return _set_norm_angles(
            self._obj,
            norm_angles=norm_angles,
            quiet=quiet,
            update_in_place=False,
            **kwargs,
        )

    def get(self):
        """
        Get the normal emission angles from the data.

        Returns
        -------
        dict
            Dictionary of normal emission angles in both `peaks` format and where applicable the true manipulator angles.

        See Also
        --------
        peaks.utils.angles._get_norm_angles

        """
        return _get_norm_angles(self._obj)


def _convert_angles_to_Ishida_Shin(data, quiet=False):
    """
    Convert the angles from friendly manipulator names into the conventions from Y. Ishida and S. Shin,
    Functions to map photoelectron distributions in a variety of setups in angle-resolved photoemission spectroscopy,
    Rev. Sci. Instrum. 89, 043903 (2018), taking care of the sign conventionsi.

    Parameters
    ----------
    data : xarray.DataArray
        Data for converting to k-space.
    quiet : bool, optional
        Whether to suppress warnings. Defaults to False.

    Returns
    -------
    dict
        Angles of converted angles.
    """

    # Get angle conventions and analyser type
    conventions = _get_conventions(data)
    ana_type = conventions.get("ana_type")

    # Get the raw angles from the data
    angles = _get_raw_angles(data, ANGLE_NAMES, quiet=quiet)

    # Adjust analyser type if necessary
    if ana_type in ["Ip", "IIp"]:
        if np.all(angles.get("defl_par", 0) == 0) and np.all(
            angles.get("defl_perp", 0) == 0
        ):
            ana_type = "I" if ana_type == "Ip" else "II"

    # Put angles in correct notation cf. Ishida and Shin
    angles_out = dict()
    angles_out["ana_type"] = ana_type
    angles_out["delta"] = angles.get("azi", 0) * conventions.get("azi", 1)
    if ana_type == "I":  # Type I
        angles_out["alpha"] = angles.get("theta_par", 0) * conventions.get(
            "theta_par", 1
        )
        angles_out["beta"] = (angles.get("polar", 0) * conventions.get("polar", 1)) + (
            angles.get("ana_polar", 0) * conventions.get("ana_polar", 1)
        )
        angles_out["xi"] = angles.get("tilt", 0) * conventions.get("tilt", 1)
    elif ana_type == "II":  # Type II
        angles_out["alpha"] = angles.get("theta_par", 0) * conventions.get(
            "theta_par", 1
        )
        angles_out["beta"] = angles.get("tilt", 0) * conventions.get("tilt", 1)
        angles_out["xi"] = (angles.get("polar", 0) * conventions.get("polar", 1)) + (
            angles.get("ana_polar", 0) * conventions.get("ana_polar", 1)
        )
    elif ana_type in ["Ip", "IIp"]:  # Type I' or Type II'
        angles_out["alpha"] = angles.get("theta_par", 0) * conventions.get(
            "theta_par", 1
        ) + (angles.get("defl_par", 0) * conventions.get("defl_par", 1))
        angles_out["beta"] = angles.get("defl_perp", 0) * conventions.get("defl_perp", 1)
        angles_out["xi"] = angles.get("tilt", 0) * conventions.get("tilt", 1)
        angles_out["chi"] = (angles.get("polar", 0) * conventions.get("polar", 1)) + (
            angles.get("ana_polar", 0) * conventions.get("ana_polar", 1)
        )

    return angles_out


def _get_angles_for_k_conv(data, return_raw=False, quiet=False, warn_norm=True):
    """
    Get the angles for the k-space conversion.

    Parameters
    ----------
    data : xarray.DataArray
        Data for converting to k-space.
    return_raw : bool, optional
        Whether to return the raw angles (polar, norm_tilt etc.), or the angles in the convention for
        k-conversion (alpha, beta etc.). Defaults to False.
    quiet : bool, optional
        Whether to suppress warnings. Defaults to False.
    warn_norm : bool, optional
        Whether to warn if normal emission angles are assumed. Defaults to True.

    Returns
    -------
    dict
        Angles for the k-space conversion (if return_raw=False) or raw angles (if return_raw=True).
    """

    # Get angle conventions and analyser type
    conventions = _get_conventions(data)
    ana_type = conventions.get("ana_type")

    # Get the raw angles from the data
    angles = _get_raw_angles(data, ANGLE_NAMES, quiet=quiet)

    # Extract normal emission angles
    norm_angles = data.norm_angles.get()

    # Make sensible guesses for missing normal emission angles
    warn_str = ""
    if warn_norm:
        for norm_angle in NORM_ANGLE_NAMES:
            if norm_angles.get(norm_angle) is None:
                if norm_angle == "norm_azi":
                    norm_angles[norm_angle] = angles.get("azi", 0) * conventions.get(
                        "azi", 1
                    )
                    warn_str += f"{norm_angle}: {norm_angles[norm_angle]}, "
                elif norm_angle == "norm_polar" and ana_type in ["I", "Ip"]:
                    norm_angles[norm_angle] = angles.get("polar", 0) * conventions.get(
                        "polar", 1
                    )
                    warn_str += f"{norm_angle}: {norm_angles[norm_angle]}, "
                elif norm_angle == "norm_tilt" and ana_type in ["II", "IIp"]:
                    norm_angles[norm_angle] = angles.get("tilt", 0) * conventions.get(
                        "tilt", 1
                    )
                    warn_str += f"{norm_angle}: {norm_angles[norm_angle]}, "
                else:
                    norm_angles[norm_angle] = 0
                    warn_str += f"{norm_angle}: 0, "
    if warn_str:
        _warn_str = f"Some normal emission data was missing or could not be passed. Assuming default values of: {warn_str.rstrip(', ')}."
        analysis_warning(_warn_str, "warning", "Analysis warning", quiet)

    # Combine angles
    angles.update(norm_angles)

    if return_raw:
        return angles

    # Adjust analyser type if necessary
    if ana_type in ["Ip", "IIp"]:
        if np.all(angles.get("defl_par", 0) == 0) and np.all(
            angles.get("defl_perp", 0) == 0
        ):
            ana_type = "I" if ana_type == "Ip" else "II"

    # Put angles in correct notation cf. Ishida and Shin, Rev. Sci. Instrum. 89 (2018) 043903
    angles_out = dict()
    angles_out["ana_type"] = ana_type
    angles_out["delta_"] = angles.get("azi", 0) * conventions.get("azi", 1) - angles.get(
        "norm_azi", 0
    )
    if ana_type == "I":  # Type I
        angles_out["alpha"] = angles.get("theta_par", 0) * conventions.get(
            "theta_par", 1
        )
        angles_out["beta"] = (angles.get("polar", 0) * conventions.get("polar", 1)) + (
            angles.get("ana_polar", 0) * conventions.get("ana_polar", 1)
        )
        angles_out["beta_0"] = angles.get("norm_polar", 0)
        print(angles_out["beta_0"])
        angles_out["xi"] = angles.get("tilt", 0) * conventions.get("tilt", 1)
        angles_out["xi_0"] = angles.get("norm_tilt", 0) * conventions.get("tilt", 1)
    elif ana_type == "II":  # Type II
        angles_out["alpha"] = angles.get("theta_par", 0) * conventions.get(
            "theta_par", 1
        )
        angles_out["beta"] = angles.get("tilt", 0) * conventions.get("tilt", 1)
        angles_out["beta_0"] = angles.get("norm_tilt", 0)
        angles_out["xi"] = (angles.get("polar", 0) * conventions.get("polar", 1)) + (
            angles.get("ana_polar", 0) * conventions.get("ana_polar", 1)
        )
        angles_out["xi_0"] = angles.get("norm_polar", 0)
    elif ana_type in ["Ip", "IIp"]:  # Type I' or Type II'
        angles_out["alpha"] = angles.get("theta_par", 0) * conventions.get(
            "theta_par", 1
        ) + (angles.get("defl_par", 0) * conventions.get("defl_par", 1))
        angles_out["beta"] = angles.get("defl_perp", 0) * conventions.get("defl_perp", 1)
        angles_out["beta_0"] = None
        angles_out["xi"] = angles.get("tilt", 0) * conventions.get("tilt", 1)
        angles_out["xi_0"] = angles.get("norm_tilt", 0)
        angles_out["chi"] = (angles.get("polar", 0) * conventions.get("polar", 1)) + (
            angles.get("ana_polar", 0) * conventions.get("ana_polar", 1)
        )
        angles_out["chi_0"] = angles.get("norm_polar", 0)

    # Add the normal emission angle sign
    angles_out["theta_par_sign"] = conventions.get("theta_par", 1)

    return angles_out



"""Common physical constants."""

# Phil King 15/05/2021
# Brendan Edwards 16/10/2023


class consts(object):
    """Class used to store constants that may be used within the peaks package.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        # Get the value of the Boltzmann constant
        kB_value = consts.kB

    """

    kvac_const = (
        0.5123167243227325  # sqrt(2m_e/hbar^2) in units for energy in eV and k in 1/AA
    )
    Cu_Ka_lambda = 1.5406  # Cu K-alpha wavelength in units of Angstroms
    kB = 1.380649 * 10**-23  # Boltzmann constant in units of J/K
    electron_volt = 1.602176634 * 10**-19  # 1 electronvolt in units of J
    electron_mass = 9.1093837 * 10**-31  # Electron mass in units of kg
    Planck_const = 6.62607015 * 10**-34  # Planck constant in units of Js
    reduced_Planck_const = (
        1.054571817 * 10**-34
    )  # Reduced Planck constant in units of Js



import pint_xarray
import xarray as xr

from peaks.core.accessors.accessor_methods import (
    _pass_function_to_xarray_class_accessor,
)

ureg = pint_xarray.unit_registry


@xr.register_datatree_accessor("tr")
@xr.register_dataarray_accessor("tr")
class TRAccessors:
    """Helper functions for time-resolved data, accessed via the `.tr` accessor."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    mean = _pass_function_to_xarray_class_accessor(
        "mean", "peaks.time_resolved.data_select"
    )
    static = _pass_function_to_xarray_class_accessor(
        "static", "peaks.time_resolved.data_select"
    )
    diff = _pass_function_to_xarray_class_accessor(
        "diff", "peaks.time_resolved.data_select"
    )
    set_t0 = _pass_function_to_xarray_class_accessor(
        "set_t0", "peaks.time_resolved.utils"
    )
    set_t0_like = _pass_function_to_xarray_class_accessor(
        "set_t0_like", "peaks.time_resolved.utils"
    )
    assign_t0 = _pass_function_to_xarray_class_accessor(
        "assign_t0", "peaks.time_resolved.utils"
    )
    assign_t0_like = _pass_function_to_xarray_class_accessor(
        "assign_t0_like", "peaks.time_resolved.utils"
    )



# ruff: noqa: I001
# ruff: noqa: F401

# Import xarray Class-based accessors
from peaks.core.metadata.history import History
from peaks.core.metadata.metadata_methods import Metadata
from peaks.core.GUI.iplot.hvplot import HVPlotAccessor
from peaks.core.fitting.quick_fit import QuickFit

# Import direct accessor methods
from peaks.core.accessors.dataarray_accessors import *  # noqa: F403
from peaks.core.accessors.dataset_accessors import *  # noqa: F403
from peaks.core.accessors.datatree_accessors import *  # noqa: F403



"""Functions for the 2D interactive display panel."""

import sys
from functools import partial

import numpy as np
import pint
import pyperclip
import pyqtgraph as pg
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QWidget,
)

from peaks.core.GUI.GUI_utils import (
    CircularListWidget,
    Crosshair,
    KeyPressGraphicsLayoutWidget,
)
from peaks.core.GUI.GUI_utils.cursor_stats import _parse_norm_emission_cursor_stats
from peaks.core.metadata.metadata_methods import display_metadata
from peaks.core.process.tools import estimate_sym_point, sym


def _disp_2d(data, primary_dim, exclude_from_centering):
    """Display a 2D interactive display panel.

    Parameters
    ------------
    data : list or xarray.DataArray
         Either a single 2D :class:`xarray.DataArray` or a list of 2D :class:`xarray.DataArray` objects.

    primary_dim : str
        The primary dimension for the viewer, will be shown on the y-axis.

    exclude_from_centering : str or tuple of str or list of str or None
        The dimension to exclude from centering. Default is 'eV'.

    """
    global app  # Ensure the QApplication instance is not garbage collected
    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)

    viewer = _Disp2D(data, primary_dim, exclude_from_centering)
    viewer.show()
    app.exec()


class _Disp2D(QtWidgets.QMainWindow):
    def __init__(self, data, primary_dim, exclude_from_centering):
        super().__init__()

        # Parse data
        if primary_dim:  # Set correct primary axis
            data = [array.transpose(primary_dim, ...) for array in data]
        self.data_arrays = [array.compute().pint.dequantify() for array in data]

        # Initialize some parameters
        self.init = False
        self.connected_plot_signals = []

        # Crosshair options
        self.num_xhs = 3
        self.start_active_xhs = 2
        self.xh_width_store = [None, None]  # Store for (dim0, dim1) DC widths
        self.xh_pos_store = [
            [None for i in range(self.num_xhs)],
            [None for i in range(self.num_xhs)],
        ]  # Store for (dim0, dim1) DC positions
        self.xh_visible_store = []
        self.DC_pens = [(238, 119, 51), (51, 187, 238), (0, 153, 136), (204, 51, 153)]
        self.xh_brushes = [
            (238, 119, 51, 70),
            (51, 187, 238, 70),
            (0, 153, 136, 70),
            (204, 51, 153, 70),
        ]

        # Set keys for the keyboard control
        def _get_key(param):
            return getattr(QtCore.Qt.Key, f"Key_{param}")

        self.key_modifiers_characters = {
            "move_csr2": "Control",
            "move_all": "Shift",
            "hide_all": "Space",
        }
        self.key_modifiers = {
            k: _get_key(v) for k, v in self.key_modifiers_characters.items()
        }
        self.show_hide_csr_key_characters = [str(i) for i in range(self.num_xhs)]
        self.show_hide_csr_keys = [
            _get_key(key) for key in self.show_hide_csr_key_characters
        ]
        self.show_hide_mirror_key_character = "M"
        self.show_hide_mirror_key = _get_key(self.show_hide_mirror_key_character)
        self.move_csr1_key_enabled = False
        self.move_all_key_enabled = False

        # Set options for dims to exclude from centering
        if isinstance(exclude_from_centering, str):
            self.exclude_from_centering = [exclude_from_centering]
        elif isinstance(exclude_from_centering, (tuple, list)):
            self.exclude_from_centering = list(exclude_from_centering)
        else:
            self.exclude_from_centering = []

        # Set up the GUI
        self._init_UI()  # Initialize the layout
        self._change_data()  # Initialize with data
        # Connect file change signals
        self.scans_list.currentRowChanged.connect(self._change_data)

        # Ensure the application quits when the window is closed
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        self.closeEvent = self._close_application

    def _close_application(self, event):
        """Close the application when the window is closed."""
        self.graphics_layout.close()
        app.quit()

    # ##############################
    # GUI layout
    # ##############################
    def _init_UI(self):
        self.setWindowTitle("Display Panel")
        self.setGeometry(100, 100, 650, 770)

        # Main window layout
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        layout = QVBoxLayout()
        central_widget.setLayout(layout)

        self._build_plot_layout(layout)  # Build the basic plot layout
        self._build_controls_layout(layout)  # Build the control panel layout
        self._build_menu()  # Add the menu

    def _build_plot_layout(self, layout):
        # Create a GraphicsLayoutWidget
        self.graphics_layout = (
            KeyPressGraphicsLayoutWidget._KeyPressGraphicsLayoutWidget()
        )
        self.graphics_layout.viewport().setAttribute(
            QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, False
        )
        layout.addWidget(self.graphics_layout)

        self.image_plot = self.graphics_layout.addPlot(1, 0, 1, 1)
        self.DC_plots = []
        self.DC_plots.append(
            self.graphics_layout.addPlot(1, 1, 1, 1)
        )  # DC_plots[0] --> EDC for our conventional data order
        self.DC_plots[0].setYLink(self.image_plot)
        self.DC_plots[0].getAxis("left").hide()
        self.DC_plots[0].getAxis("right").show()
        self.DC_plots.append(
            self.graphics_layout.addPlot(2, 0, 1, 1)
        )  # DC_plots[1] --> MDC for our conventional data order
        self.DC_plots[1].setXLink(self.image_plot)
        self.graphics_layout.ci.layout.setRowStretchFactor(1, 3)
        self.graphics_layout.ci.layout.setColumnStretchFactor(0, 3)
        # Set margins to line up the plots
        self.DC_plots[0].getAxis("right").setWidth(60)
        self.DC_plots[1].getAxis("left").setWidth(40)
        self.image_plot.getAxis("left").setWidth(40)

    def _build_controls_layout(self, layout):
        # Left panel -------------------------------------
        bottom_panel_layout = QHBoxLayout()
        layout.addLayout(bottom_panel_layout)
        bottom_panel_layout_left = QVBoxLayout()
        bottom_panel_layout_left.setSpacing(5)
        bottom_panel_layout.addLayout(bottom_panel_layout_left)

        # Create a scroll area for cursor_stats
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFixedHeight(150)
        self.cursor_stats = QLabel()
        self.cursor_stats.setStyleSheet(
            "QLabel { background-color : black; color : white; }"
        )
        self.cursor_stats.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignTop | QtCore.Qt.AlignmentFlag.AlignLeft
        )
        self.cursor_stats.setWordWrap(True)
        self.cursor_stats.setMaximumWidth(450)
        scroll_area.setWidget(self.cursor_stats)
        bottom_panel_layout_left.addWidget(scroll_area)

        # Cursor integration controls
        cursor_integrations = QHBoxLayout()
        bottom_panel_layout_left.addLayout(cursor_integrations)
        self.DC_groups = []
        self.DC_width_selectors = []
        self.DC_span_all_checkboxes = []
        DC_layouts = []
        for i in range(2):
            self.DC_groups.append(QtWidgets.QGroupBox(f"DC{i}"))
            self.DC_groups[i].setContentsMargins(5, 22, 5, 5)
            DC_layouts.append(QtWidgets.QHBoxLayout())
            DC_layouts[i].setContentsMargins(5, 0, 5, 0)
            self.DC_groups[i].setLayout(DC_layouts[i])
            cursor_integrations.addWidget(self.DC_groups[i])
            self.DC_width_selectors.append(QtWidgets.QDoubleSpinBox())
            self.DC_width_selectors[i].setDecimals(3)
            self.DC_width_selectors[i].setFixedWidth(75)
            DC_layouts[i].addWidget(self.DC_width_selectors[i])
            self.DC_span_all_checkboxes.append(QtWidgets.QCheckBox("All?"))
            DC_layouts[i].addWidget(self.DC_span_all_checkboxes[i])
        cursor_integrations.addStretch()

        # DC display controls
        DC_display_controls = QHBoxLayout()
        bottom_panel_layout_left.addLayout(DC_display_controls)
        show_DCs_group = QtWidgets.QGroupBox("Show DCs:")
        show_DCs_group.setContentsMargins(5, 22, 5, 5)
        show_DCs_layout = QtWidgets.QHBoxLayout()
        show_DCs_layout.setContentsMargins(5, 0, 5, 0)
        show_DCs_group.setLayout(show_DCs_layout)
        DC_display_controls.addWidget(show_DCs_group)
        self.show_DCs_checkboxes = []
        for i in range(self.num_xhs):
            self.show_DCs_checkboxes.append(
                QtWidgets.QCheckBox(
                    f"Csr{i}", checked=True if i < self.start_active_xhs else False
                )
            )
            show_DCs_layout.addWidget(self.show_DCs_checkboxes[i])
        self.show_mirror_checkbox = QtWidgets.QCheckBox("Mirror")
        show_DCs_layout.addWidget(self.show_mirror_checkbox)
        self.align_button = QtWidgets.QPushButton("Align")
        show_DCs_layout.addWidget(self.align_button)
        self.copy_button = QtWidgets.QPushButton("Copy")
        show_DCs_layout.addWidget(self.copy_button)

        DC_display_controls.addStretch()
        bottom_panel_layout_left.addStretch()

        # Right panel -------------------------------------
        bottom_panel_layout_right = QVBoxLayout()
        bottom_panel_layout_right.setSpacing(5)
        bottom_panel_layout.addLayout(bottom_panel_layout_right)

        # Scans list
        self.scans_list = CircularListWidget._CircularListWidget()
        self.scans_list.setFixedWidth(140)
        self.scans_list.setFixedHeight(180)
        self.scans_list.addItems([array.name for array in self.data_arrays])
        self.scans_list.setCurrentRow(0)
        bottom_panel_layout_right.addWidget(self.scans_list)

        # Add options for locking ranges on data cycle
        self.lock_opts = QtWidgets.QGroupBox("Lock?")
        self.lock_opts.setContentsMargins(5, 22, 5, 5)
        lock_opts_layout = QtWidgets.QHBoxLayout(self.lock_opts)
        lock_opts_layout.setContentsMargins(5, 0, 5, 0)
        self.cscale_lock = QtWidgets.QCheckBox("cScale")
        self.range_lock = QtWidgets.QCheckBox("Range", checked=True)
        lock_opts_layout.addWidget(self.cscale_lock)
        lock_opts_layout.addWidget(self.range_lock)
        bottom_panel_layout_right.addWidget(self.lock_opts)
        bottom_panel_layout_right.addStretch()

    def _build_menu(self):
        """Add a menu"""
        self.menu = self.menuBar().addMenu("Display Panel")
        self.shortcuts_action = QtGui.QAction("Help", self)

        self.help_text = f"""
                <table style='width:100%; border: 1px solid black; border-collapse: collapse;'>
                  <tr style='border: 1px solid black;'>
                    <th style='border: 1px solid black; padding: 5px; font-weight: normal;'>Action</th>
                    <th style='border: 1px solid black; padding: 5px; font-weight: normal;'>Key Binding</th>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Move crosshair 0</td>
                    <td style='padding: 5px; font-weight: normal;'>Arrow keys</td>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Move c_rosshair 1</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.key_modifiers_characters["move_csr2"]} + Arrow keys</td>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Move all crosshairs</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.key_modifiers_characters["move_all"]} + Arrow keys</td>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Hide all crosshairs</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.key_modifiers_characters["hide_all"]}</td>
                  </tr>
                """

        for i in range(self.num_xhs):
            self.help_text += f"""
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Enable/disable crosshair {i}</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.show_hide_csr_key_characters[i]}</td>
                  </tr>
                """

        self.help_text += f"""
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Enable/disable mirrored mode</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.show_hide_mirror_key_character}</td>
                  </tr>
                </table>
                """

        self.shortcuts_action.triggered.connect(
            lambda: QtWidgets.QMessageBox.information(self, "Help", self.help_text)
        )
        self.menu.addAction(self.shortcuts_action)

    # ##############################
    # Data handling / plotting
    # ##############################
    def _set_data(self):
        """Set the current data array and extract some core parameters."""
        self.current_data = self.data_arrays[self.scans_list.currentRow()]
        self.dims = self.current_data.dims
        self.coords = [self.current_data.coords[dim].values for dim in self.dims]
        self.step_sizes = [coord[1] - coord[0] for coord in self.coords]
        self.ranges = [
            (min(coords) - self.step_sizes[i] / 2, max(coords) + self.step_sizes[i] / 2)
            for i, coords in enumerate(self.coords)
        ]
        self.data_span = [abs(range[1] - range[0]) for range in self.ranges]
        self.c_min = float(self.current_data.min())
        self.c_max = float(self.current_data.max())

        # Set the span range box options
        for i in range(2):
            self.DC_width_selectors[i].setRange(0, self.data_span[i])
            self.DC_width_selectors[i].setSingleStep(self.step_sizes[i])

        # Set centering dims
        self.centering_dims = [
            dim for dim in self.dims if dim not in self.exclude_from_centering
        ]

        # Read scan metadata
        self.metadata_text = "<span style='color:white'>"
        self.metadata_text += display_metadata(self.current_data, "html")
        self.metadata_text += "</span><br>"

    def _set_main_plot(self, cmap, c_range, xh_pos):
        # Set main image
        self.image_item = pg.ImageItem(self.current_data.values, axisOrder="row-major")
        self.image_plot.addItem(self.image_item)

        # Transform to match data coordinates
        self.image_item.setTransform(
            pg.QtGui.QTransform(
                self.step_sizes[1],
                0,
                0,
                0,
                self.step_sizes[0],
                0,
                self.ranges[1][0],
                self.ranges[0][0],
                1,
            )
        )

        # Add colour bar
        self.colorbar = pg.ColorBarItem(
            label=self.current_data.name,
            colorMap=cmap,
            colorMapMenu=True,
            limits=(self.c_min, self.c_max),
            rounding=min(abs(self.c_max - self.c_min) / 2000, 1),
            interactive=True,
            orientation="h",
            values=c_range,
        )
        self.colorbar.setImageItem(self.image_item, insert_in=self.image_plot)
        self.image_item.setColorMap(cmap)
        self.graphics_layout.addItem(self.colorbar, row=0, col=0)

        # Add crosshairs
        xh_target_kwargs = [
            {  # Crosshair1: orange
                "pen": pg.mkPen(color=(238, 119, 51)),
                "brush": pg.mkBrush(color=(238, 119, 51, 50)),
            },
            {  # Crosshair2: blue
                "pen": pg.mkPen(color=(51, 187, 238)),
                "brush": pg.mkBrush(color=(51, 187, 238, 50)),
            },
            {  # Crosshair3: green
                "pen": pg.mkPen(color=(0, 153, 136)),
                "brush": pg.mkBrush(color=(0, 153, 136, 50)),
            },
        ]
        xh_linear_region_kwargs = [
            {  # Crosshair1: orange
                "pen": pg.mkPen(color=(238, 119, 51)),
            },
            {  # Crosshair2: blue
                "pen": pg.mkPen(color=(51, 187, 238)),
            },
            {  # Crosshair3: green
                "pen": pg.mkPen(color=(0, 153, 136)),
            },
        ]
        self.xhs = []
        for i in range(self.num_xhs):
            self.xhs.append(
                Crosshair._Crosshair(
                    self.image_plot,
                    pos=xh_pos[i],
                    dim0_width=self.DC_width_selectors[0].value(),
                    dim1_width=self.DC_width_selectors[1].value(),
                    brush=self.xh_brushes[i % len(self.xh_brushes)],
                    bounds=[self.ranges[0], self.ranges[1]],
                    axisOrder="row-major",
                    target_item_kwargs=xh_target_kwargs[i % len(xh_target_kwargs)],
                    linear_region_item_kwargs=xh_linear_region_kwargs[
                        i % len(xh_linear_region_kwargs)
                    ],
                )
            )

            # Set visibility
            self.xhs[i].set_visible(self.show_DCs_checkboxes[i].isChecked())

    def _set_DC_plots(self):
        self.DC_plot_items = {}
        self.DC_plots_xhs = [[], []]
        for dim_no in range(2):
            dim = self.dims[dim_no]
            select_along_dim_no = (dim_no + 1) % 2
            self.DC_plot_items[dim_no] = []
            self.DC_plot_items[f"{dim_no}_m"] = []
            for i in range(self.num_xhs):
                DC = self._select_DC(
                    self.current_data,
                    select_along_dim_no,
                    (
                        self.xhs[i].get_dim1_span()
                        if select_along_dim_no == 1
                        else self.xhs[i].get_dim0_span()
                    ),
                )

                # Add to plots - get ordering correct
                if dim_no == 0:
                    a, b = DC.data, DC.coords[self.dims[dim_no]].values
                else:
                    a, b = DC.coords[self.dims[dim_no]].values, DC.data

                self.DC_plot_items[dim_no].append(
                    self.DC_plots[dim_no].plot(
                        a,
                        b,
                        pen=self.DC_pens[i % len(self.DC_pens)],
                    )
                )  # Main DC

                # Set visibility
                self.DC_plot_items[dim_no][i].setVisible(
                    self.show_DCs_checkboxes[i].isChecked()
                )

                # Add a mirror DC if dim in mirror group
                if dim in self.centering_dims:
                    try:
                        mirror_DC = sym(
                            DC, flipped=True, **{dim: self.xhs[i].get_pos()[dim_no]}
                        )
                        if dim_no == 0:
                            a, b = (
                                mirror_DC.data,
                                mirror_DC.coords[self.dims[dim_no]].values,
                            )
                        else:
                            a, b = (
                                mirror_DC.coords[self.dims[dim_no]].values,
                                mirror_DC.data,
                            )
                        self.DC_plot_items[f"{dim_no}_m"].append(
                            self.DC_plots[dim_no].plot(
                                a,
                                b,
                                pen=pg.mkPen(
                                    color=self.DC_pens[i % len(self.DC_pens)],
                                    style=QtCore.Qt.PenStyle.DashLine,
                                ),
                            )
                        )  # Mirrored DC
                        self.DC_plot_items[f"{dim_no}_m"][i].setVisible(
                            self.show_mirror_checkbox.isChecked()
                        )
                    except Exception:
                        pass

            # Add spans for crosshairs
            for i in range(self.num_xhs):
                self.DC_plots_xhs[dim_no].append(
                    pg.LinearRegionItem(
                        values=getattr(self.xhs[i], f"get_dim{dim_no}_span")(),
                        orientation="horizontal" if dim_no == 0 else "vertical",
                        pen=(0, 0, 0, 0),
                        brush=self.xh_brushes[i % len(self.xh_brushes)],
                        movable=False,
                    )
                )
                self.DC_plots[dim_no].addItem(self.DC_plots_xhs[dim_no][i])
                self.DC_plots_xhs[dim_no][i].setVisible(
                    self.show_DCs_checkboxes[i].isChecked()
                )  # Set visibility

        self.DC_plots[0].getViewBox().invertX(False)

    def _set_plot_ranges(self, previous_x_range, previous_y_range):
        # Get bounds of current data
        x_min, x_max = self.ranges[1]
        y_min, y_max = self.ranges[0]

        # Check if current view is within data limits
        if (
            previous_x_range[0] > x_min
            and previous_x_range[1] < x_max
            and self.range_lock.isChecked()
        ):
            x0, x1 = previous_x_range
        else:
            x0, x1 = x_min, x_max
        if (
            previous_y_range[0] > y_min
            and previous_y_range[1] < y_max
            and self.range_lock.isChecked()
        ):
            y0, y1 = previous_y_range
        else:
            y0, y1 = y_min, y_max

        # Set ranges
        self.DC_plots[1].getViewBox().setLimits(xMin=x_min, xMax=x_max)
        self.DC_plots[1].getViewBox().setXRange(x0, x1, padding=0.0)
        self.DC_plots[0].getViewBox().setLimits(yMin=y_min, yMax=y_max)
        self.DC_plots[0].getViewBox().setYRange(y0, y1, padding=0.0)
        self.image_plot.getViewBox().setLimits(
            xMin=x_min,
            xMax=x_max,
            yMin=y_min,
            yMax=y_max,
        )
        self.image_plot.getViewBox().setRange(
            xRange=(x0, x1), yRange=(y0, y1), padding=0.0
        )

    def _set_plot_labels(self):
        # Set labels
        for i in range(2):
            dim_units = self.current_data.coords[self.dims[i]].attrs.get("units")
            dim_label = f"{self.dims[i]} ({dim_units})" if dim_units else self.dims[i]
            self.DC_plots[i].setLabel("bottom" if i == 1 else "right", dim_label)
            self.DC_groups[i].setTitle(self.dims[i])

    def _select_DC(self, data, select_along_dim_no, span_range):
        """Select a DC from the data, handling averaging."""
        select_along_dim = self.dims[select_along_dim_no]
        if np.abs(span_range[1] - span_range[0]) < self.step_sizes[select_along_dim_no]:
            return data.sel({select_along_dim: np.mean(span_range)}, method="nearest")
        else:
            return data.sel(
                {select_along_dim: slice(span_range[0], span_range[1])}
            ).mean(select_along_dim)

    # ##############################
    # Data / plot updates
    # ##############################
    def _change_data(self):
        """Change the selected data array."""
        # Disconnect signals
        for signal in self.connected_plot_signals[:]:
            signal.disconnect()
            self.connected_plot_signals.remove(signal)

        # Set data
        self._set_data()

        # Define current view ranges and crosshair positions
        xh_pos = []  # Desired crosshair positions
        if self.init and self.range_lock.isChecked():
            x_range, y_range = self.image_plot.getViewBox().viewRange()
            for i in range(self.num_xhs):
                xh_pos.append(self.xhs[i].get_pos())
                if not self._check_crosshair_in_range(xh_pos[i]):
                    xh_pos[i] = self._init_crosshair_pos(i)
        else:
            y_range, x_range = self.ranges
            for i in range(self.num_xhs):
                xh_pos.append(self._init_crosshair_pos(i))
            for i in range(2):
                self.DC_width_selectors[i].setValue(self.data_span[i] / 200)

        # Set a colour map - once set on the graph, always keep the same one
        if not self.init:
            cmap = pg.colormap.get("Greys", source="matplotlib")
        else:
            cmap = self.image_item.getColorMap()
        # Get colour range
        if self.init and self.cscale_lock.isChecked():
            c_range = self.image_item.getLevels()
        else:
            c_range = (self.c_min, self.c_max)

        # Clear existing plots
        if self.init:
            self.image_plot.clear()
            for plot in self.DC_plots:
                plot.clear()
            self.graphics_layout.removeItem(self.colorbar)

        # Make plots
        self._set_main_plot(cmap, c_range, xh_pos)  # Make main plot
        self._set_DC_plots()  # Make DC plots
        self._set_plot_ranges(x_range, y_range)  # Set ranges
        self._set_plot_labels()  # Set labels
        self._update_cursor_stats_text()  # Update cursor stats

        # Connect signals
        self._connect_signals_crosshairs()
        self._connect_signals_DCspan_change()
        self._connect_signals_align()
        self._connect_key_press_signals()

        # Mark as initialized
        self.init = True

    def _update_DC(self, xh_no):
        """Update the DC plots when the crosshair is moved."""
        # Get the xh number and related plots
        xh = self.xhs[xh_no]

        for dim_no in range(2):
            # Plot to update
            plot = self.DC_plot_items[dim_no][xh_no]

            # Get the DC
            select_along_dim_no = (dim_no + 1) % 2
            DC = self._select_DC(
                self.current_data,
                select_along_dim_no,
                (xh.get_dim1_span() if select_along_dim_no == 1 else xh.get_dim0_span()),
            )

            # Update the plot
            if dim_no == 0:
                plot.setData(DC.data, self.coords[0])
            else:
                plot.setData(self.coords[1], DC.data)

            # Check if a mirrored DC exists and needs updating
            if (
                self.dims[dim_no] in self.centering_dims
                and self.show_mirror_checkbox.isChecked()
            ):
                plot_m = self.DC_plot_items[f"{dim_no}_m"][xh_no]
                mirror_DC = sym(
                    DC, flipped=True, **{self.dims[dim_no]: xh.get_pos()[dim_no]}
                )
                if dim_no == 0:
                    plot_m.setData(
                        mirror_DC.data, mirror_DC.coords[self.dims[dim_no]].values
                    )
                else:
                    plot_m.setData(
                        mirror_DC.coords[self.dims[dim_no]].values, mirror_DC.data
                    )

    def _update_DC_width(self, dim_no):
        """Update the DC markers and plots when the width is changed."""
        for i, xh in enumerate(self.xhs):
            getattr(xh, f"set_dim{dim_no}_width")(
                self.DC_width_selectors[dim_no].value()
            )
            self.DC_plots_xhs[dim_no][i].setRegion(
                getattr(xh, f"get_dim{dim_no}_span")()
            )
            self._update_DC(i)
        self._update_cursor_stats_text()

    def _update_DC_int(self, dim_no):
        """Update the DC plots when the integration checkbox is toggled."""
        if self.DC_span_all_checkboxes[dim_no].isChecked():
            # Store current DC position and width if not already stored
            if self.xh_width_store[dim_no] is None:
                self.xh_width_store[dim_no] = self.DC_width_selectors[dim_no].value()
                self.xh_pos_store[dim_no] = [xh.get_pos()[dim_no] for xh in self.xhs]

            # Set width box to full range
            self.DC_width_selectors[dim_no].setValue(
                self.DC_width_selectors[dim_no].maximum()
            )
            self.DC_width_selectors[dim_no].setDisabled(True)

            # Force crosshair to mid point
            mid_point = np.mean(self.ranges[dim_no])
            for xh in self.xhs:
                if dim_no == 0:
                    xh.set_lock_dim0(mid_point)
                else:
                    xh.set_lock_dim1(mid_point)
                xh.update_crosshair()
        else:
            # Set to original ranges and positions
            self.DC_width_selectors[dim_no].setDisabled(False)
            self.DC_width_selectors[dim_no].setValue(self.xh_width_store[dim_no])
            self.xh_width_store[dim_no] = None
            for i, xh in enumerate(self.xhs):
                # Unlock the cursor position
                if dim_no == 0:
                    xh.set_lock_dim0(None)
                else:
                    xh.set_lock_dim1(None)
                # Reset to stored position
                current_pos = xh.get_pos()
                pos0 = (
                    self.xh_pos_store[0][i]
                    if self.xh_pos_store[0][i] is not None
                    else current_pos[0]
                )
                pos1 = (
                    self.xh_pos_store[1][i]
                    if self.xh_pos_store[1][i] is not None
                    else current_pos[1]
                )
                xh.set_pos((pos0, pos1))
                self.xh_pos_store[dim_no][i] = None
                xh.update_crosshair()

    def _update_DC_crosshair_span(self, xh_no):
        """Follow the crosshair with the DC span."""
        xh = self.xhs[xh_no]
        for i in range(2):
            self.DC_plots_xhs[i][xh_no].setRegion(getattr(xh, f"get_dim{i}_span")())

    def _show_hide_DCs(self, xh_no):
        """Show or hide the DC plots."""
        show_DC = self.show_DCs_checkboxes[xh_no].isChecked()
        show_mirror = self.show_mirror_checkbox.isChecked()
        for key, plot_item_group in self.DC_plot_items.items():
            if isinstance(key, int):
                plot_item_group[xh_no].setVisible(show_DC)
            elif isinstance(key, str) and not plot_item_group == []:
                plot_item_group[xh_no].setVisible(show_mirror and show_DC)
        self.xhs[xh_no].set_visible(show_DC)
        self.DC_plots_xhs[0][xh_no].setVisible(show_DC)
        self.DC_plots_xhs[1][xh_no].setVisible(show_DC)
        self._update_cursor_stats_text()

    def _show_hide_mirror(self):
        for i in range(self.num_xhs):
            self._show_hide_DCs(i)

    def _align_data(self):
        """Estimate symmetry points in the data"""
        # Get the current view range
        x_range, y_range = self.image_plot.getViewBox().viewRange()
        data_to_centre = self.current_data.sel(
            {self.dims[0]: slice(*y_range), self.dims[1]: slice(*x_range)}
        )

        # Get the centre of the data
        centre = estimate_sym_point(data_to_centre, dims=self.centering_dims)

        # Activate mirror mode
        self.show_mirror_checkbox.setChecked(True)

        # Set crosshairs to the aligned position
        for xh in self.xhs:
            current_pos = xh.get_pos()
            dim0_pos = centre.get(self.dims[0], current_pos[0])
            dim1_pos = centre.get(self.dims[1], current_pos[1])
            xh.set_pos((dim0_pos, dim1_pos))

    def _update_cursor_stats_text(self):
        """Update the cursor stats."""

        cursor_colors = [f"#{r:02x}{g:02x}{b:02x}" for r, g, b in self.DC_pens]
        cursor_text = ""

        # Get data from active cursors
        cursor_pos = [
            xh.get_pos() if self.show_DCs_checkboxes[i].isChecked() else None
            for i, xh in enumerate(self.xhs)
        ]
        cursor_spans0 = [
            xh.get_dim0_span() if self.show_DCs_checkboxes[i].isChecked() else None
            for i, xh in enumerate(self.xhs)
        ]
        cursor_spans1 = [
            xh.get_dim1_span() if self.show_DCs_checkboxes[i].isChecked() else None
            for i, xh in enumerate(self.xhs)
        ]

        # Show current cursor positions and values (integrated over the crosshair selection)
        cursor_values = []
        for i in range(self.num_xhs):
            if cursor_pos[i] is None:
                cursor_values.append(None)
                cursor_text += "<br>"
            else:
                span0 = cursor_spans0[i]
                span1 = cursor_spans1[i]
                if abs(span0[1] - span0[0]) < self.step_sizes[0]:
                    data = self.current_data.sel(
                        {self.dims[0]: np.mean(span0)}, method="nearest"
                    )
                else:
                    data = self.current_data.sel(
                        {self.dims[0]: slice(span0[0], span0[1])}
                    ).mean(self.dims[0])
                if abs(span1[1] - span1[0]) < self.step_sizes[1]:
                    data = data.sel({self.dims[1]: np.mean(span1)}, method="nearest")
                else:
                    data = data.sel({self.dims[1]: slice(span1[0], span1[1])}).mean(
                        self.dims[1]
                    )
                cursor_values.append(data.mean().values)
                cursor_text += (
                    f"<span style='color:{cursor_colors[i % len(cursor_colors)]}'>"
                    f"Csr{i} {self.dims[0]}: {cursor_pos[i][0]:.3f} | "
                    f"{self.dims[1]}: {cursor_pos[i][1]:.3f} | "
                    f"Value: {cursor_values[i]:.3f}</span><br>"
                )

        # Show the delta between the first two cursors if active
        if cursor_pos[0] is not None and cursor_pos[1] is not None:
            cursor_text += (
                f"<span style='color:white'>"
                f"ΔCsr1-0: Δ{self.dims[0]}: {abs(cursor_pos[1][0] - cursor_pos[0][0]):.3f} | "
                f"Δ{self.dims[1]}: {abs(cursor_pos[1][1] - cursor_pos[0][1]):.3f}</span>"
            )
        else:
            cursor_text += "<br>"

        # Add normal emission
        if cursor_pos[0] is not None:
            norm_emission = self._get_norm_values()
            if norm_emission:
                cursor_text += "<br>"
                cursor_text += _parse_norm_emission_cursor_stats(
                    self.current_data, norm_emission
                )
        # Add metadata
        cursor_text += self.metadata_text
        # Set text
        self.cursor_stats.setText(cursor_text)

    def _get_norm_values(self):
        cursor_pos = [
            xh.get_pos() if self.show_DCs_checkboxes[i].isChecked() else None
            for i, xh in enumerate(self.xhs)
        ]

        try:
            norm_values = self.current_data.metadata.get_normal_emission_from_values(
                {
                    self.dims[0]: float(cursor_pos[0][0]),
                    self.dims[1]: float(cursor_pos[0][1]),
                },
            )
        except AttributeError:
            norm_values = {}
        return norm_values

    def _copy_norm_values(self):
        """Copy the normal emission values to the clipboard."""
        # Current norm values
        norm_values = self._get_norm_values()
        norm_values_to_return = {
            k: str(v) if isinstance(v, pint.Quantity) else v
            for k, v in norm_values.items()
        }
        pyperclip.copy(f".metadata.set_normal_emission({norm_values_to_return})")

    # ##############################
    # Signal connections
    # ##############################
    def _connect_signals_crosshairs(self):
        # Update when crosshair moves
        for i, xh_ in enumerate(self.xhs):
            signal = xh_.xh.sigPositionChanged
            signal.connect(partial(self._update_DC, i))  # Update DC plots
            signal.connect(partial(self._update_DC_crosshair_span, i))  # Follow DC span
            signal.connect(self._update_cursor_stats_text)  # Update label
            self.connected_plot_signals.append(signal)

        # Update when show/hide crosshair boxes toggled
        for i, checkbox in enumerate(self.show_DCs_checkboxes):
            checkbox.stateChanged.connect(partial(self._show_hide_DCs, i))
            self.connected_plot_signals.append(checkbox.stateChanged)

        # Update when show/hide mirror checkbox toggled
        signal = self.show_mirror_checkbox.stateChanged
        signal.connect(self._show_hide_mirror)
        self.connected_plot_signals.append(signal)

    def _connect_signals_DCspan_change(self):
        # Update when span ranges changed
        for i in range(2):
            signal = self.DC_width_selectors[i].valueChanged
            signal.connect(partial(self._update_DC_width, i))
            self.connected_plot_signals.append(signal)

        # Update with integrate all selection
        for i in range(2):
            signal = self.DC_span_all_checkboxes[i].stateChanged
            signal.connect(partial(self._update_DC_int, i))
            self.connected_plot_signals.append(signal)

    def _connect_signals_align(self):
        # Connect centre data button
        self.align_button.clicked.connect(self._align_data)
        self.connected_plot_signals.append(self.align_button.clicked)
        # Connect copy button
        self.copy_button.clicked.connect(self._copy_norm_values)

    def _connect_key_press_signals(self):
        signals = [self.graphics_layout.keyPressed, self.graphics_layout.keyReleased]
        fns = [self._key_press_event, self._key_release_event]
        for signal, fn in zip(signals, fns, strict=True):
            signal.connect(fn)
            self.connected_plot_signals.append(signal)

    def _key_press_event(self, event):
        # First deal with the modifiers
        if event.key() == self.key_modifiers["move_csr2"]:
            self.move_csr1_key_enabled = True
        elif event.key() == self.key_modifiers["move_all"]:
            self.move_all_key_enabled = True
        elif event.key() == self.key_modifiers["hide_all"]:
            # If this is the first press, store what xhs are currently visible and hide all
            if len(self.xh_visible_store) == 0:
                for i in range(self.num_xhs):
                    if self.show_DCs_checkboxes[i].isChecked():
                        self.xh_visible_store.append(self.xhs[i])
                        self.xhs[i].set_visible(False)
        elif event.key() in self.show_hide_csr_keys:
            # Show/hide cursors
            csr_no = self.show_hide_csr_keys.index(event.key())
            self.show_DCs_checkboxes[csr_no].setChecked(
                not self.show_DCs_checkboxes[csr_no].isChecked()
            )
        elif event.key() == self.show_hide_mirror_key:
            self.show_mirror_checkbox.setChecked(
                not self.show_mirror_checkbox.isChecked()
            )
        elif event.key() in [
            QtCore.Qt.Key.Key_Up,
            QtCore.Qt.Key.Key_Down,
            QtCore.Qt.Key.Key_Left,
            QtCore.Qt.Key.Key_Right,
        ]:
            if self.move_all_key_enabled:
                xhs = [
                    xh
                    for i, xh in enumerate(self.xhs)
                    if self.show_DCs_checkboxes[i].isChecked()
                ]
            elif self.move_csr1_key_enabled:
                xhs = [self.xhs[1]]
            else:
                xhs = [self.xhs[0]]
            for xh in xhs:
                current_pos = xh.get_pos()
                if event.key() == QtCore.Qt.Key.Key_Right:
                    xh.set_pos((current_pos[0], current_pos[1] + self.step_sizes[1]))
                elif event.key() == QtCore.Qt.Key.Key_Left:
                    xh.set_pos((current_pos[0], current_pos[1] - self.step_sizes[1]))
                elif event.key() == QtCore.Qt.Key.Key_Up:
                    xh.set_pos((current_pos[0] + self.step_sizes[0], current_pos[1]))
                elif event.key() == QtCore.Qt.Key.Key_Down:
                    xh.set_pos((current_pos[0] - self.step_sizes[0], current_pos[1]))

    def _key_release_event(self, event):
        if event.key() == self.key_modifiers["move_csr2"]:
            self.move_csr1_key_enabled = False
        elif event.key() == self.key_modifiers["move_all"]:
            self.move_all_key_enabled = False
        elif event.key() == self.key_modifiers["hide_all"]:
            # Restore xh to previous state
            for xh in self.xh_visible_store:
                xh.set_visible(True)
            self.xh_visible_store = []

    # ##############################
    # Helper functions
    # ##############################
    def _init_crosshair_pos(self, xh_no):
        """Initialize the crosshair position."""
        active_xh = [
            self.show_DCs_checkboxes[i].isChecked() for i in range(self.num_xhs)
        ]
        num_active_xh = sum(active_xh)
        csr_no_active = sum(active_xh[:xh_no])
        percentile = min((csr_no_active + 1) * 100 / (num_active_xh + 1), 95)
        return (
            np.percentile(self.ranges[0], percentile),
            np.percentile(self.ranges[1], percentile),
        )

    def _check_crosshair_in_range(self, pos):
        """Check if crosshair pos is within the coordinate range"""
        return (min(self.ranges[0]) <= pos[0] <= max(self.ranges[0])) and (
            min(self.ranges[1]) <= pos[1] <= max(self.ranges[1])
        )
